# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Networks useful for building recurrent agents.
"""

import functools
from typing import NamedTuple, Optional, Sequence, Tuple
from absl import logging
from acme import types
from acme.tf import savers
from acme.tf import utils
from acme.tf.networks import base
import sonnet as snt
import tensorflow as tf
import tensorflow_probability as tfp
import tree

RNNState = types.NestedTensor


class PolicyCriticRNNState(NamedTuple):
  """Consists of two RNNStates called 'policy' and 'critic'."""
  policy: RNNState
  critic: RNNState


class UnpackWrapper(snt.Module):
  """Gets a list of arguments and pass them as separate arguments.

  Example
  ```
    class Critic(snt.Module):
      def __call__(self, o, a):
        pass

    critic = Critic()
    UnpackWrapper(critic)((o, a))
  ```
  calls critic(o, a)
  """

  def __init__(self, module: snt.Module, name: str = 'UnpackWrapper'):
    super().__init__(name=name)
    self._module = module

  def __call__(self,
               inputs: Sequence[types.NestedTensor]) -> types.NestedTensor:
    # Unpack the inputs before passing to the underlying module.
    return self._module(*inputs)


class RNNUnpackWrapper(snt.RNNCore):
  """Gets a list of arguments and pass them as separate arguments.

  Example
  ```
    class Critic(snt.RNNCore):
      def __call__(self, o, a, prev_state):
        pass

    critic = Critic()
    RNNUnpackWrapper(critic)((o, a), prev_state)
  ```
  calls m(o, a, prev_state)
  """

  def __init__(self, module: snt.RNNCore, name: str = 'RNNUnpackWrapper'):
    super().__init__(name=name)
    self._module = module

  def __call__(self, inputs: Sequence[types.NestedTensor],
               prev_state: RNNState) -> Tuple[types.NestedTensor, RNNState]:
    # Unpack the inputs before passing to the underlying module.
    return self._module(*inputs, prev_state)

  def initial_state(self, batch_size):
    return self._module.initial_state(batch_size)


class CriticDeepRNN(snt.DeepRNN):
  """Same as snt.DeepRNN, but takes three inputs (obs, act, prev_state).
  """

  def __init__(self, layers: Sequence[snt.Module]):
    # Make the first layer take a single input instead of a list of arguments.
    if isinstance(layers[0], snt.RNNCore):
      first_layer = RNNUnpackWrapper(layers[0])
    else:
      first_layer = UnpackWrapper(layers[0])
    super().__init__([first_layer] + list(layers[1:]))

    self._unwrapped_first_layer = layers[0]
    self.__input_signature = None

  def __call__(self, inputs: types.NestedTensor, action: tf.Tensor,
               prev_state: RNNState) -> Tuple[types.NestedTensor, RNNState]:
    # Pack the inputs into a tuple and then using inherited DeepRNN logic to
    # pass them through the layers.
    # This in turn will pass the packed inputs into the first layer
    # (UnpackWrapper) which will unpack them back.
    return super().__call__((inputs, action), prev_state)

  @property
  def _input_signature(self) -> Optional[tf.TensorSpec]:
    """Return input signature for Acme snapshotting.

    The Acme way of snapshotting works as follows: you first create your network
    variables via the utility function `acme.tf.utils.create_variables()`, which
    adds an `_input_signature` attribute to your module. This attribute is
    critical for proper snapshot saving and loading.

    If a module with such an attribute is wrapped into e.g. DeepRNN, Acme
    descends into the `_layers[0]` of that DeepRNN to find the input
    signature.

    This implementation allows CriticDeepRNN to work seamlessly like DeepRNN for
    the following two use-cases:

        1) Creating variables *before* wrapping:
          ```
          unwrapped_critic = Critic()
          acme.tf.utils.create_variables(unwrapped_critic, specs)
          critic = CriticDeepRNN([unwrapped_critic])
          ```

        2) Create variables *after* wrapping:
          ```
          unwrapped_critic = Critic()
          critic = CriticDeepRNN([unwrapped_critic])
          acme.tf.utils.create_variables(critic, specs)
          ```

    Returns:
      input_signature of the module or None of it is not known (i.e. the
      variables were not created by acme.tf.utils.create_variables nor for this
      module nor for any of its descendants).
    """

    if self.__input_signature is not None:
      # To make case (2) (see above) work, we need to allow create_variables to
      # assign an _input_signature attribute to this module, which is why we
      # create additional __input_signature attribute with a setter (see below).
      return self.__input_signature

    # To make case (1) work, we descend into self._unwrapped_first_layer
    # and try to get its input signature (if it exists) by calling
    # savers.get_input_signature.

    # Ideally, savers.get_input_signature should automatically descend into
    # DeepRNN. But in this case it breaks on CriticDeepRNN because
    # CriticDeepRNN._layers[0] is an UnpackWrapper around the underlying module
    # and not the module itself.
    input_signature = savers._get_input_signature(self._unwrapped_first_layer)  # pylint: disable=protected-access
    if input_signature is None:
      return None
    # Since adding recurrent modules via CriticDeepRNN changes the recurrent
    # state, we need to update its spec here.
    state = self.initial_state(1)
    input_signature[-1] = tree.map_structure(
        lambda t: tf.TensorSpec((None,) + t.shape[1:], t.dtype), state)
    self.__input_signature = input_signature
    return input_signature

  @_input_signature.setter
  def _input_signature(self, new_spec: tf.TensorSpec):
    self.__input_signature = new_spec


class RecurrentExpQWeightedPolicy(snt.RNNCore):
  """Recurrent exponentially Q-weighted policy."""

  def __init__(self,
               policy_network: snt.Module,
               critic_network: snt.Module,
               temperature_beta: float = 1.0,
               num_action_samples: int = 16):
    super().__init__(name='RecurrentExpQWeightedPolicy')
    self._policy_network = policy_network
    self._critic_network = critic_network
    self._num_action_samples = num_action_samples
    self._temperature_beta = temperature_beta

  def __call__(self,
               observation: types.NestedTensor,
               prev_state: PolicyCriticRNNState
               ) -> Tuple[types.NestedTensor, PolicyCriticRNNState]:

    return tf.vectorized_map(self._call, (observation, prev_state))

  def _call(
      self, observation_and_state: Tuple[types.NestedTensor,
                                         PolicyCriticRNNState]
  ) -> Tuple[types.NestedTensor, PolicyCriticRNNState]:
    """Computes a forward step for a single element.

    The observation and state are packed together in order to use
    `tf.vectorized_map` to handle batches of observations.
    See this module's __call__() function.

    Args:
      observation_and_state: the observation and state packed in a tuple.

    Returns:
      The selected action and the corresponding state.
    """
    observation, prev_state = observation_and_state

    # Tile input observations and states to allow multiple policy predictions.
    tiled_observation, tiled_prev_state = utils.tile_nested(
        (observation, prev_state), self._num_action_samples)
    actions, policy_states = self._policy_network(
        tiled_observation, tiled_prev_state.policy)

    # Evaluate multiple critic predictions with the sampled actions.
    value_distribution, critic_states = self._critic_network(
        tiled_observation, actions, tiled_prev_state.critic)
    value_estimate = value_distribution.mean()

    # Resample a single action of the sampled actions according to logits given
    # by the tempered Q-values.
    selected_action_idx = tfp.distributions.Categorical(
        probs=tf.nn.softmax(value_estimate / self._temperature_beta)).sample()
    selected_action = actions[selected_action_idx]

    # Select and return the RNN state that corresponds to the selected action.
    states = PolicyCriticRNNState(
        policy=policy_states, critic=critic_states)
    selected_state = tree.map_structure(
        lambda x: x[selected_action_idx], states)

    return selected_action, selected_state

  def initial_state(self, batch_size: int) -> PolicyCriticRNNState:
    return PolicyCriticRNNState(
        policy=self._policy_network.initial_state(batch_size),
        critic=self._critic_network.initial_state(batch_size)
        )


class DeepRNN(snt.DeepRNN, base.RNNCore):
  """Unroll-aware deep RNN module.

  Sonnet's DeepRNN steps through RNNCores sequentially which can result in a
  performance hit, in particular when using Transformers. This module adds an
  unroll() method which unrolls each module in the DeepRNN individually,
  allowing efficient implementation of the unroll operation. For example, a
  Transformer can 'unroll' by evaluating the whole sequence at once (this being
  one of the advantages of Transformers over e.g. LSTMs).

  Any RNNCore passed to this module should implement unroll(). Failure to so
  may cause the RNNCore not to be called properly. For example, passing a
  partial function application of a snt.RNNCore to this module will fail (this
  is also true for snt.DeepRNN). However, the special case of passing in a
  RNNCore object that does not implement unroll() is supported and will be
  dynamically unrolled. Implement unroll() to override this behavior with
  static unrolling.

  Stateless modules (i.e. anything other than an RNNCore) which do not
  implement unroll() are batch applied over the time and batch axes
  simultaneously. Effectively, this means that such modules may be applied to
  fairly large batches, potentially leading to out-of-memory issues.
  """

  def __init__(self, layers, name: Optional[str] = None):
    """Initializes the module."""
    super().__init__(layers, name=name)

    self.__input_signature = None
    self._num_unrollable = 0

    # As a convenience, check for snt.RNNCore modules and dynamically unroll
    # them if they don't already support unrolling. This check can fail, e.g.
    # if a partially applied RNNCore is passed in. Sonnet's implementation of
    # DeepRNN suffers from the same problem.
    for layer in self._layers:
      if hasattr(layer, 'unroll'):
        self._num_unrollable += 1
      elif isinstance(layer, snt.RNNCore):
        self._num_unrollable += 1
        layer.unroll = functools.partial(snt.dynamic_unroll, layer)
        logging.warning(
            'Acme DeepRNN detected a Sonnet RNNCore. '
            'This will be dynamically unrolled. Please implement unroll() '
            'to suppress this warning.')

  def unroll(self,
             inputs: types.NestedTensor,
             state: base.State,
             sequence_length: int,
             ) -> Tuple[types.NestedTensor, base.State]:
    """Unroll each layer individually.

    Calls unroll() on layers which support it, all other layers are
    batch-applied over the first two axes (assumed to be the time and batch
    axes).

    Args:
      inputs: A nest of `tf.Tensor` in time-major format.
      state: The RNN core state.
      sequence_length: How long the static_unroll should go for.

    Returns:
      Nested sequence output of RNN, and final state.

    Raises:
      ValueError if the length of `state` does not match the number of
      unrollable layers.
    """
    if len(state) != self._num_unrollable:
      raise ValueError(
          'DeepRNN was called with the wrong number of states. The length of '
          '`state` does not match the number of unrollable layers.')

    states = iter(state)
    outputs = inputs
    next_states = []
    for layer in self._layers:
      if hasattr(layer, 'unroll'):
        # The length of the `states` list was checked above.
        outputs, next_state = layer.unroll(outputs, next(states),
                                           sequence_length)
        next_states.append(next_state)
      else:
        # Couldn't unroll(); assume that this is a stateless module.
        outputs = snt.BatchApply(layer, num_dims=2)(outputs)

    return outputs, tuple(next_states)

  @property
  def _input_signature(self) -> Optional[tf.TensorSpec]:
    """Return input signature for Acme snapshotting, see CriticDeepRNN."""

    if self.__input_signature is not None:
      return self.__input_signature

    input_signature = savers._get_input_signature(self._layers[0])  # pylint: disable=protected-access
    if input_signature is None:
      return None

    state = self.initial_state(1)
    input_signature[-1] = tree.map_structure(
        lambda t: tf.TensorSpec((None,) + t.shape[1:], t.dtype), state)
    self.__input_signature = input_signature
    return input_signature

  @_input_signature.setter
  def _input_signature(self, new_spec: tf.TensorSpec):
    self.__input_signature = new_spec


class LSTM(snt.LSTM, base.RNNCore):
  """Unrollable interface to LSTM.

  This module is supposed to be used with the DeepRNN class above, and more
  generally in networks which support unroll().
  """

  def unroll(self,
             inputs: types.NestedTensor,
             state: base.State,
             sequence_length: int,
             ) -> Tuple[types.NestedTensor, base.State]:
    return snt.static_unroll(self, inputs, state, sequence_length)



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for discrete BCQ learner."""

from absl.testing import absltest

from acme import specs
from acme.agents.tf import bcq
from acme.testing import fakes
from acme.tf import utils as tf2_utils
from acme.tf.networks import discrete as discrete_networks
from acme.utils import counting

import numpy as np
import sonnet as snt


def _make_network(action_spec: specs.DiscreteArray) -> snt.Module:
  return snt.Sequential([
      snt.Flatten(),
      snt.nets.MLP([50, 50, action_spec.num_values]),
  ])


class DiscreteBCQLearnerTest(absltest.TestCase):

  def test_full_learner(self):
    # Create dataset.
    environment = fakes.DiscreteEnvironment(
        num_actions=5,
        num_observations=10,
        obs_dtype=np.float32,
        episode_length=10)
    spec = specs.make_environment_spec(environment)
    dataset = fakes.transition_dataset(environment).batch(2)

    # Build network.
    g_network = _make_network(spec.actions)
    q_network = _make_network(spec.actions)
    network = discrete_networks.DiscreteFilteredQNetwork(g_network=g_network,
                                                         q_network=q_network,
                                                         threshold=0.5)
    tf2_utils.create_variables(network, [spec.observations])

    # Build learner.
    counter = counting.Counter()
    learner = bcq.DiscreteBCQLearner(
        network=network,
        dataset=dataset,
        learning_rate=1e-4,
        discount=0.99,
        importance_sampling_exponent=0.2,
        target_update_period=100,
        counter=counter)

    # Run a learner step.
    learner.step()

    # Check counts from BC and BCQ learners.
    counts = counter.get_counts()
    self.assertEqual(1, counts['bc_steps'])
    self.assertEqual(1, counts['bcq_steps'])

    # Check learner state.
    self.assertEqual(1, learner.state['bc_num_steps'].numpy())
    self.assertEqual(1, learner.state['bcq_num_steps'].numpy())


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common types used throughout Acme."""

from typing import Any, Callable, Iterable, Mapping, NamedTuple, Union
from acme import specs

# Define types for nested arrays and tensors.
# TODO(b/144758674): Replace these with recursive type definitions.
NestedArray = Any
NestedTensor = Any

# pytype: disable=not-supported-yet
NestedSpec = Union[
    specs.Array,
    Iterable['NestedSpec'],
    Mapping[Any, 'NestedSpec'],
]
# pytype: enable=not-supported-yet

# TODO(b/144763593): Replace all instances of nest with the tensor/array types.
Nest = Union[NestedArray, NestedTensor, NestedSpec]

TensorTransformation = Callable[[NestedTensor], NestedTensor]
TensorValuedCallable = Callable[..., NestedTensor]


class Batches(int):
  """Helper class for specification of quantities in units of batches.

  Example usage:

      # Configure the batch size and replay size in units of batches.
      config.batch_size = 32
      config.replay_size = Batches(4)

      # ...

      # Convert the replay size at runtime.
      if isinstance(config.replay_size, Batches):
        config.replay_size = config.replay_size * config.batch_size  # int: 128

  """


class Transition(NamedTuple):
  """Container for a transition."""
  observation: NestedArray
  action: NestedArray
  reward: NestedArray
  discount: NestedArray
  next_observation: NestedArray
  extras: NestedArray = ()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Frame stacking utilities."""

import collections

from acme import types
from acme.wrappers import base
import dm_env
from dm_env import specs as dm_env_specs
import numpy as np
import tree


class FrameStackingWrapper(base.EnvironmentWrapper):
  """Wrapper that stacks observations along a new final axis."""

  def __init__(self, environment: dm_env.Environment, num_frames: int = 4):
    """Initializes a new FrameStackingWrapper.

    Args:
      environment: Environment.
      num_frames: Number frames to stack.
    """
    self._environment = environment
    original_spec = self._environment.observation_spec()
    self._stackers = tree.map_structure(
        lambda _: FrameStacker(num_frames=num_frames),
        self._environment.observation_spec())
    self._observation_spec = tree.map_structure(
        lambda stacker, spec: stacker.update_spec(spec),
        self._stackers, original_spec)

  def _process_timestep(self, timestep: dm_env.TimeStep) -> dm_env.TimeStep:
    observation = tree.map_structure(lambda stacker, x: stacker.step(x),
                                     self._stackers, timestep.observation)
    return timestep._replace(observation=observation)

  def reset(self) -> dm_env.TimeStep:
    for stacker in tree.flatten(self._stackers):
      stacker.reset()
    return self._process_timestep(self._environment.reset())

  def step(self, action: int) -> dm_env.TimeStep:
    return self._process_timestep(self._environment.step(action))

  def observation_spec(self) -> types.NestedSpec:
    return self._observation_spec


class FrameStacker:
  """Simple class for frame-stacking observations."""

  def __init__(self, num_frames: int):
    self._num_frames = num_frames
    self.reset()

  @property
  def num_frames(self) -> int:
    return self._num_frames

  def reset(self):
    self._stack = collections.deque(maxlen=self._num_frames)

  def step(self, frame: np.ndarray) -> np.ndarray:
    if not self._stack:
      # Fill stack with blank frames if empty.
      self._stack.extend([np.zeros_like(frame)] * (self._num_frames - 1))
    self._stack.append(frame)
    return np.stack(self._stack, axis=-1)

  def update_spec(self, spec: dm_env_specs.Array) -> dm_env_specs.Array:
    return dm_env_specs.Array(shape=spec.shape + (self._num_frames,),
                              dtype=spec.dtype,
                              name=spec.name)



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wrapping trfl epsilon_greedy with legal action masking."""

from typing import Optional, Mapping, Union

import sonnet as snt
import tensorflow as tf
import trfl


class NetworkWithMaskedEpsilonGreedy(snt.Module):
  """Epsilon greedy sampling with action masking on network outputs."""

  def __init__(self,
               network: snt.Module,
               epsilon: Optional[tf.Tensor] = None):
    """Initialize the network and epsilon.

    Usage:
      Wrap an observation in a dictionary in your environment as follows:

        observation <-- {"your_key_for_observation": observation,
                         "legal_actions_mask": your_action_mask_tensor}

    and update your network to use 'observation["your_key_for_observation"]'
    rather than 'observation'.

    Args:
      network: the online Q network (the one being optimized)
      epsilon: probability of taking a random action.
    """
    super().__init__()
    self._network = network
    self._epsilon = epsilon

  def __call__(
      self, observation: Union[Mapping[str, tf.Tensor],
                               tf.Tensor]) -> tf.Tensor:
    q = self._network(observation)
    return trfl.epsilon_greedy(
        q, epsilon=self._epsilon,
        legal_actions_mask=observation['legal_actions_mask']).sample()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TD3 config."""
import dataclasses
from typing import Optional, Union

from acme.adders import reverb as adders_reverb
import optax


@dataclasses.dataclass
class TD3Config:
  """Configuration options for TD3."""

  # Loss options
  batch_size: int = 256
  policy_learning_rate: Union[optax.Schedule, float] = 3e-4
  critic_learning_rate: Union[optax.Schedule, float] = 3e-4
  # Policy gradient clipping is not part of the original TD3 implementation,
  # used e.g. in DAC https://arxiv.org/pdf/1809.02925.pdf
  policy_gradient_clipping: Optional[float] = None
  discount: float = 0.99
  n_step: int = 1

  # TD3 specific options (https://arxiv.org/pdf/1802.09477.pdf)
  sigma: float = 0.1
  delay: int = 2
  target_sigma: float = 0.2
  noise_clip: float = 0.5
  tau: float = 0.005

  # Replay options
  min_replay_size: int = 1000
  max_replay_size: int = 1000000
  replay_table_name: str = adders_reverb.DEFAULT_PRIORITY_TABLE
  prefetch_size: int = 4
  samples_per_insert: float = 256
  # Rate to be used for the SampleToInsertRatio rate limiter tolerance.
  # See a formula in make_replay_tables for more details.
  samples_per_insert_tolerance_rate: float = 0.1

  # How many gradient updates to perform per step.
  num_sgd_steps_per_step: int = 1



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements the MPO loss.

The MPO loss uses MPOParams, which can be initialized using init_params,
to track the temperature and the dual variables.

Tensor shapes are annotated, where helpful, as follow:
  B: batch size,
  N: number of sampled actions, see MPO paper for more details,
  D: dimensionality of the action space.
"""

from typing import NamedTuple, Optional, Tuple, Union

import jax
import jax.numpy as jnp
import tensorflow_probability

tfp = tensorflow_probability.substrates.jax
tfd = tensorflow_probability.substrates.jax.distributions

_MPO_FLOAT_EPSILON = 1e-8

Shape = Tuple[int]
DType = type(jnp.float32)  # _ScalarMeta, a private type.


class MPOParams(NamedTuple):
  """NamedTuple to store trainable loss parameters."""
  log_temperature: jnp.ndarray
  log_alpha_mean: jnp.ndarray
  log_alpha_stddev: jnp.ndarray
  log_penalty_temperature: Optional[jnp.ndarray] = None


class MPOStats(NamedTuple):
  """NamedTuple to store loss statistics."""
  dual_alpha_mean: float
  dual_alpha_stddev: float
  dual_temperature: float

  loss_policy: float
  loss_alpha: float
  loss_temperature: float
  kl_q_rel: float

  kl_mean_rel: float
  kl_stddev_rel: float

  q_min: float
  q_max: float

  pi_stddev_min: float
  pi_stddev_max: float
  pi_stddev_cond: float

  penalty_kl_q_rel: Optional[float] = None


class MPO:
  """MPO loss with decoupled KL constraints as in (Abdolmaleki et al., 2018).

  This implementation of the MPO loss includes the following features, as
  options:
  - Satisfying the KL-constraint on a per-dimension basis (on by default);
  - Penalizing actions that fall outside of [-1, 1] (on by default) as a
      special case of multi-objective MPO (MO-MPO; Abdolmaleki et al., 2020).
  For best results on the control suite, keep both of these on.

  (Abdolmaleki et al., 2018): https://arxiv.org/pdf/1812.02256.pdf
  (Abdolmaleki et al., 2020): https://arxiv.org/pdf/2005.07513.pdf
  """

  def __init__(self,
               epsilon: float,
               epsilon_mean: float,
               epsilon_stddev: float,
               init_log_temperature: float,
               init_log_alpha_mean: float,
               init_log_alpha_stddev: float,
               per_dim_constraining: bool = True,
               action_penalization: bool = True,
               epsilon_penalty: float = 0.001):
    """Initialize and configure the MPO loss.

    Args:
      epsilon: KL constraint on the non-parametric auxiliary policy, the one
        associated with the dual variable called temperature.
      epsilon_mean: KL constraint on the mean of the Gaussian policy, the one
        associated with the dual variable called alpha_mean.
      epsilon_stddev: KL constraint on the stddev of the Gaussian policy, the
        one associated with the dual variable called alpha_mean.
      init_log_temperature: initial value for the temperature in log-space, note
        a softplus (rather than an exp) will be used to transform this.
      init_log_alpha_mean: initial value for the alpha_mean in log-space, note a
        softplus (rather than an exp) will be used to transform this.
      init_log_alpha_stddev: initial value for the alpha_stddev in log-space,
        note a softplus (rather than an exp) will be used to transform this.
      per_dim_constraining: whether to enforce the KL constraint on each
        dimension independently; this is the default. Otherwise the overall KL
        is constrained, which allows some dimensions to change more at the
        expense of others staying put.
      action_penalization: whether to use a KL constraint to penalize actions
        via the MO-MPO algorithm.
      epsilon_penalty: KL constraint on the probability of violating the action
        constraint.
    """

    # MPO constrain thresholds.
    self._epsilon = epsilon
    self._epsilon_mean = epsilon_mean
    self._epsilon_stddev = epsilon_stddev

    # Initial values for the constraints' dual variables.
    self._init_log_temperature = init_log_temperature
    self._init_log_alpha_mean = init_log_alpha_mean
    self._init_log_alpha_stddev = init_log_alpha_stddev

    # Whether to penalize out-of-bound actions via MO-MPO and its corresponding
    # constraint threshold.
    self._action_penalization = action_penalization
    self._epsilon_penalty = epsilon_penalty

    # Whether to ensure per-dimension KL constraint satisfication.
    self._per_dim_constraining = per_dim_constraining

  def init_params(self, action_dim: int, dtype: DType = jnp.float32):
    """Creates an initial set of parameters."""

    if self._per_dim_constraining:
      dual_variable_shape = [action_dim]
    else:
      dual_variable_shape = [1]

    log_temperature = jnp.full([1], self._init_log_temperature, dtype=dtype)

    log_alpha_mean = jnp.full(
        dual_variable_shape, self._init_log_alpha_mean, dtype=dtype)

    log_alpha_stddev = jnp.full(
        dual_variable_shape, self._init_log_alpha_stddev, dtype=dtype)

    if self._action_penalization:
      log_penalty_temperature = jnp.full([1],
                                         self._init_log_temperature,
                                         dtype=dtype)
    else:
      log_penalty_temperature = None

    return MPOParams(
        log_temperature=log_temperature,
        log_alpha_mean=log_alpha_mean,
        log_alpha_stddev=log_alpha_stddev,
        log_penalty_temperature=log_penalty_temperature)

  def __call__(
      self,
      params: MPOParams,
      online_action_distribution: Union[tfd.MultivariateNormalDiag,
                                        tfd.Independent],
      target_action_distribution: Union[tfd.MultivariateNormalDiag,
                                        tfd.Independent],
      actions: jnp.ndarray,  # Shape [N, B, D].
      q_values: jnp.ndarray,  # Shape [N, B].
  ) -> Tuple[jnp.ndarray, MPOStats, MPOParams]:
    """Computes the decoupled MPO loss.

    Args:
      params: parameters tracking the temperature and the dual variables.
      online_action_distribution: online distribution returned by the online
        policy network; expects batch_dims of [B] and event_dims of [D].
      target_action_distribution: target distribution returned by the target
        policy network; expects same shapes as online distribution.
      actions: actions sampled from the target policy; expects shape [N, B, D].
      q_values: Q-values associated with each action; expects shape [N, B].

    Returns:
      Loss, combining the policy loss, KL penalty, and dual losses required to
        adapt the dual variables.
      Stats, for diagnostics and tracking performance.
    """

    # Cast `MultivariateNormalDiag`s to Independent Normals.
    # The latter allows us to satisfy KL constraints per-dimension.
    if isinstance(target_action_distribution, tfd.MultivariateNormalDiag):
      target_action_distribution = tfd.Independent(
          tfd.Normal(target_action_distribution.mean(),
                     target_action_distribution.stddev()))
      online_action_distribution = tfd.Independent(
          tfd.Normal(online_action_distribution.mean(),
                     online_action_distribution.stddev()))

    # Infer the shape and dtype of dual variables.
    scalar_dtype = q_values.dtype

    # Create dual variables for the KL constraints; only happens the first call.
    # self.create_dual_variables_once(dual_variable_shape, scalar_dtype)

    # Project dual variables to ensure they stay positive.
    min_log_temperature = jnp.array(-18.0, scalar_dtype)
    min_log_alpha = jnp.array(-18.0, scalar_dtype)

    log_temperature = params.log_temperature
    log_temperature = jnp.maximum(min_log_temperature, log_temperature)

    log_alpha_mean = params.log_alpha_mean
    log_alpha_mean = jnp.maximum(min_log_alpha, log_alpha_mean)

    log_alpha_stddev = params.log_alpha_stddev
    log_alpha_stddev = jnp.maximum(min_log_alpha, log_alpha_stddev)

    # Transform dual variables from log-space.
    # Note: using softplus instead of exponential for numerical stability.
    temperature = jax.nn.softplus(log_temperature) + _MPO_FLOAT_EPSILON
    alpha_mean = jax.nn.softplus(log_alpha_mean) + _MPO_FLOAT_EPSILON
    alpha_stddev = jax.nn.softplus(log_alpha_stddev) + _MPO_FLOAT_EPSILON

    # Get online and target means and stddevs in preparation for decomposition.
    online_mean = online_action_distribution.distribution.mean()
    online_scale = online_action_distribution.distribution.stddev()
    target_mean = target_action_distribution.distribution.mean()
    target_scale = target_action_distribution.distribution.stddev()

    # Compute normalized importance weights, used to compute expectations with
    # respect to the non-parametric policy; and the temperature loss, used to
    # adapt the tempering of Q-values.
    normalized_weights, loss_temperature = compute_weights_and_temperature_loss(
        q_values, self._epsilon, temperature)

    # Only needed for diagnostics: Compute estimated actualized KL between the
    # non-parametric and current target policies.
    kl_nonparametric = compute_nonparametric_kl_from_normalized_weights(
        normalized_weights)

    if self._action_penalization:
      # Project and transform action penalization temperature.

      log_penalty_temperature = params.log_penalty_temperature
      log_penalty_temperature = jnp.maximum(min_log_temperature,
                                            log_penalty_temperature)

      penalty_temperature = jax.nn.softplus(
          log_penalty_temperature) + _MPO_FLOAT_EPSILON

      # Compute action penalization cost.
      # Note: the cost is zero in [-1, 1] and quadratic beyond.
      diff_out_of_bound = actions - jnp.clip(actions, -1.0, 1.0)
      cost_out_of_bound = -jnp.linalg.norm(diff_out_of_bound, axis=-1)

      penalty_normalized_weights, loss_penalty_temperature = compute_weights_and_temperature_loss(
          cost_out_of_bound, self._epsilon_penalty, penalty_temperature)

      # Only needed for diagnostics: Compute estimated actualized KL between the
      # non-parametric and current target policies.
      penalty_kl_nonparametric = compute_nonparametric_kl_from_normalized_weights(
          penalty_normalized_weights)

      # Combine normalized weights.
      normalized_weights += penalty_normalized_weights
      loss_temperature += loss_penalty_temperature
    else:
      log_penalty_temperature = None

    # Decompose the online policy into fixed-mean & fixed-stddev distributions.
    # This has been documented as having better performance in bandit settings,
    # see e.g. https://arxiv.org/pdf/1812.02256.pdf.
    fixed_stddev_distribution = tfd.Independent(
        tfd.Normal(loc=online_mean, scale=target_scale))
    fixed_mean_distribution = tfd.Independent(
        tfd.Normal(loc=target_mean, scale=online_scale))

    # Compute the decomposed policy losses.
    loss_policy_mean = compute_cross_entropy_loss(actions, normalized_weights,
                                                  fixed_stddev_distribution)
    loss_policy_stddev = compute_cross_entropy_loss(actions, normalized_weights,
                                                    fixed_mean_distribution)

    # Compute the decomposed KL between the target and online policies.
    if self._per_dim_constraining:
      kl_mean = target_action_distribution.distribution.kl_divergence(
          fixed_stddev_distribution.distribution)  # Shape [B, D].
      kl_stddev = target_action_distribution.distribution.kl_divergence(
          fixed_mean_distribution.distribution)  # Shape [B, D].
    else:
      kl_mean = target_action_distribution.kl_divergence(
          fixed_stddev_distribution)  # Shape [B].
      kl_stddev = target_action_distribution.kl_divergence(
          fixed_mean_distribution)  # Shape [B].

    # Compute the alpha-weighted KL-penalty and dual losses to adapt the alphas.
    loss_kl_mean, loss_alpha_mean = compute_parametric_kl_penalty_and_dual_loss(
        kl_mean, alpha_mean, self._epsilon_mean)
    loss_kl_stddev, loss_alpha_stddev = compute_parametric_kl_penalty_and_dual_loss(
        kl_stddev, alpha_stddev, self._epsilon_stddev)

    # Combine losses.
    loss_policy = loss_policy_mean + loss_policy_stddev
    loss_kl_penalty = loss_kl_mean + loss_kl_stddev
    loss_dual = loss_alpha_mean + loss_alpha_stddev + loss_temperature
    loss = loss_policy + loss_kl_penalty + loss_dual

    # Create statistics.
    pi_stddev = online_action_distribution.distribution.stddev()
    stats = MPOStats(
        # Dual Variables.
        dual_alpha_mean=jnp.mean(alpha_mean),
        dual_alpha_stddev=jnp.mean(alpha_stddev),
        dual_temperature=jnp.mean(temperature),
        # Losses.
        loss_policy=jnp.mean(loss),
        loss_alpha=jnp.mean(loss_alpha_mean + loss_alpha_stddev),
        loss_temperature=jnp.mean(loss_temperature),
        # KL measurements.
        kl_q_rel=jnp.mean(kl_nonparametric) / self._epsilon,
        penalty_kl_q_rel=((jnp.mean(penalty_kl_nonparametric) /
                           self._epsilon_penalty)
                          if self._action_penalization else None),
        kl_mean_rel=jnp.mean(kl_mean, axis=0) / self._epsilon_mean,
        kl_stddev_rel=jnp.mean(kl_stddev, axis=0) / self._epsilon_stddev,
        # Q measurements.
        q_min=jnp.mean(jnp.min(q_values, axis=0)),
        q_max=jnp.mean(jnp.max(q_values, axis=0)),
        # If the policy has stddev, log summary stats for this as well.
        pi_stddev_min=jnp.mean(jnp.min(pi_stddev, axis=-1)),
        pi_stddev_max=jnp.mean(jnp.max(pi_stddev, axis=-1)),
        # Condition number of the diagonal covariance (actually, stddev) matrix.
        pi_stddev_cond=jnp.mean(
            jnp.max(pi_stddev, axis=-1) / jnp.min(pi_stddev, axis=-1)),
    )

    # Create new state
    state_new = MPOParams(
        log_temperature=log_temperature,
        log_alpha_mean=log_alpha_mean,
        log_alpha_stddev=log_alpha_stddev,
        log_penalty_temperature=log_penalty_temperature)

    return loss, stats, state_new


def compute_weights_and_temperature_loss(
    q_values: jnp.ndarray,
    epsilon: float,
    temperature: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
  """Computes normalized importance weights for the policy optimization.

  Args:
    q_values: Q-values associated with the actions sampled from the target
      policy; expected shape [N, B].
    epsilon: Desired constraint on the KL between the target and non-parametric
      policies.
    temperature: Scalar used to temper the Q-values before computing normalized
      importance weights from them. This is really the Lagrange dual variable in
      the constrained optimization problem, the solution of which is the
      non-parametric policy targeted by the policy loss.

  Returns:
    Normalized importance weights, used for policy optimization.
    Temperature loss, used to adapt the temperature.
  """

  # Temper the given Q-values using the current temperature.
  tempered_q_values = jax.lax.stop_gradient(q_values) / temperature

  # Compute the normalized importance weights used to compute expectations with
  # respect to the non-parametric policy.
  normalized_weights = jax.nn.softmax(tempered_q_values, axis=0)
  normalized_weights = jax.lax.stop_gradient(normalized_weights)

  # Compute the temperature loss (dual of the E-step optimization problem).
  q_logsumexp = jax.scipy.special.logsumexp(tempered_q_values, axis=0)
  log_num_actions = jnp.log(q_values.shape[0] / 1.)
  loss_temperature = epsilon + jnp.mean(q_logsumexp) - log_num_actions
  loss_temperature = temperature * loss_temperature

  return normalized_weights, loss_temperature


def compute_nonparametric_kl_from_normalized_weights(
    normalized_weights: jnp.ndarray) -> jnp.ndarray:
  """Estimate the actualized KL between the non-parametric and target policies."""

  # Compute integrand.
  num_action_samples = normalized_weights.shape[0] / 1.
  integrand = jnp.log(num_action_samples * normalized_weights + 1e-8)

  # Return the expectation with respect to the non-parametric policy.
  return jnp.sum(normalized_weights * integrand, axis=0)


def compute_cross_entropy_loss(
    sampled_actions: jnp.ndarray,
    normalized_weights: jnp.ndarray,
    online_action_distribution: tfd.Distribution,
) -> jnp.ndarray:
  """Compute cross-entropy online and the reweighted target policy.

  Args:
    sampled_actions: samples used in the Monte Carlo integration in the policy
      loss. Expected shape is [N, B, ...], where N is the number of sampled
      actions and B is the number of sampled states.
    normalized_weights: target policy multiplied by the exponentiated Q values
      and normalized; expected shape is [N, B].
    online_action_distribution: policy to be optimized.

  Returns:
    loss_policy_gradient: the cross-entropy loss that, when differentiated,
      produces the policy gradient.
  """

  # Compute the M-step loss.
  log_prob = online_action_distribution.log_prob(sampled_actions)

  # Compute the weighted average log-prob using the normalized weights.
  loss_policy_gradient = -jnp.sum(log_prob * normalized_weights, axis=0)

  # Return the mean loss over the batch of states.
  return jnp.mean(loss_policy_gradient, axis=0)


def compute_parametric_kl_penalty_and_dual_loss(
    kl: jnp.ndarray,
    alpha: jnp.ndarray,
    epsilon: float,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
  """Computes the KL cost to be added to the Lagragian and its dual loss.

  The KL cost is simply the alpha-weighted KL divergence and it is added as a
  regularizer to the policy loss. The dual variable alpha itself has a loss that
  can be minimized to adapt the strength of the regularizer to keep the KL
  between consecutive updates at the desired target value of epsilon.

  Args:
    kl: KL divergence between the target and online policies.
    alpha: Lagrange multipliers (dual variables) for the KL constraints.
    epsilon: Desired value for the KL.

  Returns:
    loss_kl: alpha-weighted KL regularization to be added to the policy loss.
    loss_alpha: The Lagrange dual loss minimized to adapt alpha.
  """

  # Compute the mean KL over the batch.
  mean_kl = jnp.mean(kl, axis=0)

  # Compute the regularization.
  loss_kl = jnp.sum(jax.lax.stop_gradient(alpha) * mean_kl)

  # Compute the dual loss.
  loss_alpha = jnp.sum(alpha * (epsilon - jax.lax.stop_gradient(mean_kl)))

  return loss_kl, loss_alpha



# Lint as: python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test networks for building recurrent agents."""

import os

from absl.testing import absltest
from acme import specs
from acme.tf import savers as tf2_savers
from acme.tf import utils as tf2_utils
from acme.tf.networks import recurrence

import numpy as np
import sonnet as snt
import tensorflow as tf
import tree


# Simple critic-like modules for testing.
class Critic(snt.Module):

  def __call__(self, o, a):
    return o * a


class RNNCritic(snt.RNNCore):

  def __call__(self, o, a, prev_state):
    return o * a, prev_state

  def initial_state(self, batch_size):
    return ()


class NetsTest(tf.test.TestCase):

  def test_criticdeeprnn_snapshot(self):
    """Test that CriticDeepRNN works correctly with snapshotting."""
    # Create a test network.
    critic = Critic()
    rnn_critic = RNNCritic()

    for base_net in [critic, rnn_critic]:
      net = recurrence.CriticDeepRNN([base_net, snt.LSTM(10)])
      obs = specs.Array([10], dtype=np.float32)
      actions = specs.Array([10], dtype=np.float32)
      spec = [obs, actions]
      tf2_utils.create_variables(net, spec)

      # Test that if you add some postprocessing without rerunning
      # create_variables, it still works.
      wrapped_net = recurrence.CriticDeepRNN([net, lambda x: x])

      for curr_net in [net, wrapped_net]:
        # Save the test network.
        directory = absltest.get_default_test_tmpdir()
        objects_to_save = {'net': curr_net}
        snapshotter = tf2_savers.Snapshotter(
            objects_to_save, directory=directory)
        snapshotter.save()

        # Reload the test network.
        net2 = tf.saved_model.load(os.path.join(snapshotter.directory, 'net'))

        obs = tf.ones((2, 10))
        actions = tf.ones((2, 10))
        state = curr_net.initial_state(2)
        outputs1, next_state1 = curr_net(obs, actions, state)
        outputs2, next_state2 = net2(obs, actions, state)

        assert np.allclose(outputs1, outputs2)
        assert np.allclose(tree.flatten(next_state1), tree.flatten(next_state2))


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for utils."""

from absl.testing import absltest
from acme.jax import utils

import chex
import jax
import jax.numpy as jnp
import numpy as np

chex.set_n_cpu_devices(4)


class JaxUtilsTest(absltest.TestCase):

  def test_batch_concat(self):
    batch_size = 32
    inputs = [
        jnp.zeros(shape=(batch_size, 2)),
        {
            'foo': jnp.zeros(shape=(batch_size, 5, 3))
        },
        [jnp.zeros(shape=(batch_size, 1))],
        jnp.zeros(shape=(batch_size,)),
    ]

    output_shape = utils.batch_concat(inputs).shape
    expected_shape = [batch_size, 2 + 5 * 3 + 1 + 1]
    self.assertSequenceEqual(output_shape, expected_shape)

  def test_mapreduce(self):

    @utils.mapreduce
    def f(y, x):
      return jnp.square(x + y)

    z = f(jnp.ones(shape=(32,)), jnp.ones(shape=(32,)))
    z = jax.device_get(z)
    self.assertEqual(z, 4)

  def test_get_from_first_device(self):
    sharded = {
        'a':
            jax.device_put_sharded(
                list(jnp.arange(16).reshape([jax.local_device_count(), 4])),
                jax.local_devices()),
        'b':
            jax.device_put_sharded(
                list(jnp.arange(8).reshape([jax.local_device_count(), 2])),
                jax.local_devices(),
            ),
    }

    want = {
        'a': jnp.arange(4),
        'b': jnp.arange(2),
    }

    # Get zeroth device content as DeviceArray.
    device_arrays = utils.get_from_first_device(sharded, as_numpy=False)
    jax.tree_map(
        lambda x: self.assertIsInstance(x, jax.xla.DeviceArray),
        device_arrays)
    jax.tree_map(np.testing.assert_array_equal, want, device_arrays)

    # Get the zeroth device content as numpy arrays.
    numpy_arrays = utils.get_from_first_device(sharded, as_numpy=True)
    jax.tree_map(lambda x: self.assertIsInstance(x, np.ndarray), numpy_arrays)
    jax.tree_map(np.testing.assert_array_equal, want, numpy_arrays)

  def test_get_from_first_device_fails_if_sda_not_provided(self):
    with self.assertRaises(ValueError):
      utils.get_from_first_device({'a': np.arange(jax.local_device_count())})


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wraps an OpenSpiel RL environment to be used as a dm_env environment."""

from typing import List, NamedTuple

from acme import specs
from acme import types
import dm_env
import numpy as np
# pytype: disable=import-error
from open_spiel.python import rl_environment
# pytype: enable=import-error


class OLT(NamedTuple):
  """Container for (observation, legal_actions, terminal) tuples."""
  observation: types.Nest
  legal_actions: types.Nest
  terminal: types.Nest


class OpenSpielWrapper(dm_env.Environment):
  """Environment wrapper for OpenSpiel RL environments."""

  # Note: we don't inherit from base.EnvironmentWrapper because that class
  # assumes that the wrapped environment is a dm_env.Environment.

  def __init__(self, environment: rl_environment.Environment):
    self._environment = environment
    self._reset_next_step = True
    if not environment.is_turn_based:
      raise ValueError("Currently only supports turn based games.")

  def reset(self) -> dm_env.TimeStep:
    """Resets the episode."""
    self._reset_next_step = False
    open_spiel_timestep = self._environment.reset()
    observations = self._convert_observation(open_spiel_timestep)
    return dm_env.restart(observations)

  def step(self, action: types.NestedArray) -> dm_env.TimeStep:
    """Steps the environment."""
    if self._reset_next_step:
      return self.reset()

    open_spiel_timestep = self._environment.step(action)

    if open_spiel_timestep.step_type == rl_environment.StepType.LAST:
      self._reset_next_step = True

    observations = self._convert_observation(open_spiel_timestep)
    rewards = np.asarray(open_spiel_timestep.rewards)
    discounts = np.asarray(open_spiel_timestep.discounts)
    step_type = open_spiel_timestep.step_type

    if step_type == rl_environment.StepType.FIRST:
      step_type = dm_env.StepType.FIRST
    elif step_type == rl_environment.StepType.MID:
      step_type = dm_env.StepType.MID
    elif step_type == rl_environment.StepType.LAST:
      step_type = dm_env.StepType.LAST
    else:
      raise ValueError(
          "Did not recognize OpenSpiel StepType: {}".format(step_type))

    return dm_env.TimeStep(observation=observations,
                           reward=rewards,
                           discount=discounts,
                           step_type=step_type)

  # Convert OpenSpiel observation so it's dm_env compatible. Also, the list
  # of legal actions must be converted to a legal actions mask.
  def _convert_observation(
      self, open_spiel_timestep: rl_environment.TimeStep) -> List[OLT]:
    observations = []
    for pid in range(self._environment.num_players):
      legals = np.zeros(self._environment.game.num_distinct_actions(),
                        dtype=np.float32)
      legals[open_spiel_timestep.observations["legal_actions"][pid]] = 1.0
      player_observation = OLT(observation=np.asarray(
          open_spiel_timestep.observations["info_state"][pid],
          dtype=np.float32),
                               legal_actions=legals,
                               terminal=np.asarray([open_spiel_timestep.last()],
                                                   dtype=np.float32))
      observations.append(player_observation)
    return observations

  def observation_spec(self) -> OLT:
    # Observation spec depends on whether the OpenSpiel environment is using
    # observation/information_state tensors.
    if self._environment.use_observation:
      return OLT(observation=specs.Array(
          (self._environment.game.observation_tensor_size(),), np.float32),
                 legal_actions=specs.Array(
                     (self._environment.game.num_distinct_actions(),),
                     np.float32),
                 terminal=specs.Array((1,), np.float32))
    else:
      return OLT(observation=specs.Array(
          (self._environment.game.information_state_tensor_size(),),
          np.float32),
                 legal_actions=specs.Array(
                     (self._environment.game.num_distinct_actions(),),
                     np.float32),
                 terminal=specs.Array((1,), np.float32))

  def action_spec(self) -> specs.DiscreteArray:
    return specs.DiscreteArray(self._environment.game.num_distinct_actions())

  def reward_spec(self) -> specs.BoundedArray:
    return specs.BoundedArray((),
                              np.float32,
                              minimum=self._environment.game.min_utility(),
                              maximum=self._environment.game.max_utility())

  def discount_spec(self) -> specs.BoundedArray:
    return specs.BoundedArray((), np.float32, minimum=0, maximum=1.0)

  @property
  def environment(self) -> rl_environment.Environment:
    """Returns the wrapped environment."""
    return self._environment

  @property
  def current_player(self) -> int:
    return self._environment.get_state.current_player()

  def __getattr__(self, name: str):
    """Expose any other attributes of the underlying environment."""
    return getattr(self._environment, name)



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TD3 learner implementation."""

import time
from typing import Dict, Iterator, List, NamedTuple, Optional, Tuple

import acme
from acme import types
from acme.agents.jax.td3 import networks as td3_networks
from acme.jax import networks as networks_lib
from acme.jax import utils
from acme.utils import counting
from acme.utils import loggers
import jax
import jax.numpy as jnp
import optax
import reverb
import rlax


class TrainingState(NamedTuple):
  """Contains training state for the learner."""
  policy_params: networks_lib.Params
  target_policy_params: networks_lib.Params
  critic_params: networks_lib.Params
  target_critic_params: networks_lib.Params
  twin_critic_params: networks_lib.Params
  target_twin_critic_params: networks_lib.Params
  policy_opt_state: optax.OptState
  critic_opt_state: optax.OptState
  twin_critic_opt_state: optax.OptState
  steps: int
  random_key: networks_lib.PRNGKey


class TD3Learner(acme.Learner):
  """TD3 learner."""

  _state: TrainingState

  def __init__(self,
               networks: td3_networks.TD3Networks,
               random_key: networks_lib.PRNGKey,
               discount: float,
               iterator: Iterator[reverb.ReplaySample],
               policy_optimizer: optax.GradientTransformation,
               critic_optimizer: optax.GradientTransformation,
               twin_critic_optimizer: optax.GradientTransformation,
               delay: int = 2,
               target_sigma: float = 0.2,
               noise_clip: float = 0.5,
               tau: float = 0.005,
               counter: Optional[counting.Counter] = None,
               logger: Optional[loggers.Logger] = None,
               num_sgd_steps_per_step: int = 1):
    """Initializes the TD3 learner.

    Args:
      networks: TD3 networks.
      random_key: a key for random number generation.
      discount: discount to use for TD updates
      iterator: an iterator over training data.
      policy_optimizer: the policy optimizer.
      critic_optimizer: the Q-function optimizer.
      twin_critic_optimizer: the twin Q-function optimizer.
      delay: ratio of policy updates for critic updates (see TD3),
        delay=2 means 2 updates of the critic for 1 policy update.
      target_sigma: std of zero mean Gaussian added to the action of
        the next_state, for critic evaluation (reducing overestimation bias).
      noise_clip: hard constraint on target noise.
      tau: target parameters smoothing coefficient.
      counter: counter object used to keep track of steps.
      logger: logger object to be used by learner.
      num_sgd_steps_per_step: number of sgd steps to perform per learner 'step'.
    """

    def policy_loss(
        policy_params: networks_lib.Params,
        critic_params: networks_lib.Params,
        observation: types.NestedArray,
    ) -> jnp.ndarray:
      # Computes the discrete policy gradient loss.
      action = networks.policy_network.apply(policy_params, observation)
      grad_critic = jax.vmap(
          jax.grad(networks.critic_network.apply, argnums=2),
          in_axes=(None, 0, 0))
      dq_da = grad_critic(critic_params, observation, action)
      batch_dpg_learning = jax.vmap(rlax.dpg_loss, in_axes=(0, 0))
      loss = batch_dpg_learning(action, dq_da)
      return jnp.mean(loss)

    def critic_loss(
        critic_params: networks_lib.Params,
        state: TrainingState,
        transition: types.Transition,
        random_key: jnp.ndarray,
    ):
      # Computes the critic loss.
      q_tm1 = networks.critic_network.apply(
          critic_params, transition.observation, transition.action)

      action = networks.policy_network.apply(state.target_policy_params,
                                             transition.next_observation)
      action_with_noise = networks.add_policy_noise(action, random_key,
                                                    target_sigma, noise_clip)

      q_t = networks.critic_network.apply(
          state.target_critic_params,
          transition.next_observation,
          action_with_noise)
      twin_q_t = networks.twin_critic_network.apply(
          state.target_twin_critic_params,
          transition.next_observation,
          action_with_noise)

      q_t = jnp.minimum(q_t, twin_q_t)

      target_q_tm1 = transition.reward + discount * transition.discount * q_t
      td_error = jax.lax.stop_gradient(target_q_tm1) - q_tm1

      return jnp.mean(jnp.square(td_error))

    def update_step(
        state: TrainingState,
        transitions: types.Transition,
    ) -> Tuple[TrainingState, Dict[str, jnp.ndarray]]:

      random_key, key_critic, key_twin = jax.random.split(state.random_key, 3)
      polyak_averaging = lambda x, y: x * (1 - tau) + y * tau

      # Updates on the critic: compute the gradients, and update using
      # Polyak averaging.
      critic_loss_and_grad = jax.value_and_grad(critic_loss)
      critic_loss_value, critic_gradients = critic_loss_and_grad(
          state.critic_params, state, transitions, key_critic)
      critic_updates, critic_opt_state = critic_optimizer.update(
          critic_gradients, state.critic_opt_state)
      critic_params = optax.apply_updates(state.critic_params, critic_updates)
      target_critic_params = jax.tree_multimap(
          polyak_averaging, state.target_critic_params, critic_params)

      # Updates on the twin critic: compute the gradients, and update using
      # Polyak averaging.
      twin_critic_loss_value, twin_critic_gradients = critic_loss_and_grad(
          state.twin_critic_params, state, transitions, key_twin)
      twin_critic_updates, twin_critic_opt_state = twin_critic_optimizer.update(
          twin_critic_gradients, state.twin_critic_opt_state)
      twin_critic_params = optax.apply_updates(state.twin_critic_params,
                                               twin_critic_updates)
      target_twin_critic_params = jax.tree_multimap(
          polyak_averaging, state.target_twin_critic_params, twin_critic_params)

      # Updates on the policy: compute the gradients, and update using
      # Polyak averaging (if delay enabled, the update might not be applied).
      policy_loss_and_grad = jax.value_and_grad(policy_loss)
      policy_loss_value, policy_gradients = policy_loss_and_grad(
          state.policy_params, state.critic_params,
          transitions.next_observation)
      def update_policy_step():
        policy_updates, policy_opt_state = policy_optimizer.update(
            policy_gradients, state.policy_opt_state)
        policy_params = optax.apply_updates(state.policy_params, policy_updates)
        target_policy_params = jax.tree_multimap(
            polyak_averaging, state.target_policy_params, policy_params)
        return policy_params, target_policy_params, policy_opt_state

      # The update on the policy is applied every `delay` steps.
      current_policy_state = (state.policy_params, state.target_policy_params,
                              state.policy_opt_state)
      policy_params, target_policy_params, policy_opt_state = jax.lax.cond(
          state.steps % delay == 0,
          lambda _: update_policy_step(),
          lambda _: current_policy_state,
          operand=None)

      steps = state.steps + 1

      new_state = TrainingState(
          policy_params=policy_params,
          critic_params=critic_params,
          twin_critic_params=twin_critic_params,
          target_policy_params=target_policy_params,
          target_critic_params=target_critic_params,
          target_twin_critic_params=target_twin_critic_params,
          policy_opt_state=policy_opt_state,
          critic_opt_state=critic_opt_state,
          twin_critic_opt_state=twin_critic_opt_state,
          steps=steps,
          random_key=random_key,
      )

      metrics = {
          'policy_loss': policy_loss_value,
          'critic_loss': critic_loss_value,
          'twin_critic_loss': twin_critic_loss_value,
      }

      return new_state, metrics

    # General learner book-keeping and loggers.
    self._counter = counter or counting.Counter()
    self._logger = logger or loggers.make_default_logger(
        'learner', asynchronous=True, serialize_fn=utils.fetch_devicearray)

    # Create prefetching dataset iterator.
    self._iterator = iterator

    # Faster sgd step
    update_step = utils.process_multiple_batches(update_step,
                                                 num_sgd_steps_per_step)
    # Use the JIT compiler.
    self._update_step = jax.jit(update_step)

    (key_init_policy, key_init_twin, key_init_target, key_state
     ) = jax.random.split(random_key, 4)
    # Create the network parameters and copy into the target network parameters.
    initial_policy_params = networks.policy_network.init(key_init_policy)
    initial_critic_params = networks.critic_network.init(key_init_twin)
    initial_twin_critic_params = networks.twin_critic_network.init(
        key_init_target)

    initial_target_policy_params = initial_policy_params
    initial_target_critic_params = initial_critic_params
    initial_target_twin_critic_params = initial_twin_critic_params

    # Initialize optimizers.
    initial_policy_opt_state = policy_optimizer.init(initial_policy_params)
    initial_critic_opt_state = critic_optimizer.init(initial_critic_params)
    initial_twin_critic_opt_state = twin_critic_optimizer.init(
        initial_twin_critic_params)

    # Create initial state.
    self._state = TrainingState(
        policy_params=initial_policy_params,
        target_policy_params=initial_target_policy_params,
        critic_params=initial_critic_params,
        twin_critic_params=initial_twin_critic_params,
        target_critic_params=initial_target_critic_params,
        target_twin_critic_params=initial_target_twin_critic_params,
        policy_opt_state=initial_policy_opt_state,
        critic_opt_state=initial_critic_opt_state,
        twin_critic_opt_state=initial_twin_critic_opt_state,
        steps=0,
        random_key=key_state
    )

    # Do not record timestamps until after the first learning step is done.
    # This is to avoid including the time it takes for actors to come online and
    # fill the replay buffer.
    self._timestamp = None

  def step(self):
    # Get data from replay (dropping extras if any). Note there is no
    # extra data here because we do not insert any into Reverb.
    sample = next(self._iterator)
    transitions = types.Transition(*sample.data)

    self._state, metrics = self._update_step(self._state, transitions)

    # Compute elapsed time.
    timestamp = time.time()
    elapsed_time = timestamp - self._timestamp if self._timestamp else 0
    self._timestamp = timestamp

    # Increment counts and record the current time
    counts = self._counter.increment(steps=1, walltime=elapsed_time)

    # Attempts to write the logs.
    self._logger.write({**metrics, **counts})

  def get_variables(self, names: List[str]) -> List[networks_lib.Params]:
    variables = {
        'policy': self._state.policy_params,
        'critic': self._state.critic_params,
        'twin_critic': self._state.twin_critic_params,
    }
    return [variables[name] for name in names]

  def save(self) -> TrainingState:
    return self._state

  def restore(self, state: TrainingState):
    self._state = state



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""RL agent Builder interface."""

import abc
from typing import Any, Generic, Iterator, List, Optional

from acme import adders
from acme import core
from acme import specs
from acme.jax import networks as networks_lib
from acme.jax.types import Networks, PolicyNetwork, Sample  # pylint: disable=g-multiple-import
from acme.utils import counting
import reverb


class GenericActorLearnerBuilder(abc.ABC, Generic[Networks, PolicyNetwork,
                                                  Sample]):
  """Defines an interface for defining the components of an RL agent.

  Implementations of this interface contain a complete specification of a
  concrete RL agent. An instance of this class can be used to build an
  RL agent which interacts with the environment either locally or in a
  distributed setup.
  """

  @abc.abstractmethod
  def make_replay_tables(
      self,
      environment_spec: specs.EnvironmentSpec,
  ) -> List[reverb.Table]:
    """Create tables to insert data into."""

  @abc.abstractmethod
  def make_dataset_iterator(
      self,
      replay_client: reverb.Client,
  ) -> Iterator[Sample]:
    """Create a dataset iterator to use for learning/updating the agent."""

  @abc.abstractmethod
  def make_adder(
      self,
      replay_client: reverb.Client,
  ) -> Optional[adders.Adder]:
    """Create an adder which records data generated by the actor/environment.

    Args:
      replay_client: Reverb Client which points to the replay server.
    """

  @abc.abstractmethod
  def make_actor(
      self,
      random_key: networks_lib.PRNGKey,
      policy_network: PolicyNetwork,
      adder: Optional[adders.Adder] = None,
      variable_source: Optional[core.VariableSource] = None,
  ) -> core.Actor:
    """Create an actor instance.

    Args:
      random_key: A key for random number generation.
      policy_network: Instance of a policy network; this should be a callable
        which takes as input observations and returns actions.
      adder: How data is recorded (e.g. added to replay).
      variable_source: A source providing the necessary actor parameters.
    """

  @abc.abstractmethod
  def make_learner(
      self,
      random_key: networks_lib.PRNGKey,
      networks: Networks,
      dataset: Iterator[Sample],
      replay_client: Optional[reverb.Client] = None,
      counter: Optional[counting.Counter] = None,
  ) -> core.Learner:
    """Creates an instance of the learner.

    Args:
      random_key: A key for random number generation.
      networks: struct describing the networks needed by the learner; this can
        be specific to the learner in question.
      dataset: iterator over samples from replay.
      replay_client: client which allows communication with replay, e.g. in
        order to update priorities.
      counter: a Counter which allows for recording of counts (learner steps,
        actor steps, etc.) distributed throughout the agent.
    """


class ActorLearnerBuilder(GenericActorLearnerBuilder[Any, Any,
                                                     reverb.ReplaySample]):
  pass



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper methods for handling signals."""

import contextlib
import ctypes
import threading
from typing import Any, Callable, Optional

import launchpad

_Handler = Callable[[], Any]


@contextlib.contextmanager
def runtime_terminator(callback: Optional[_Handler] = None):
  """Runtime terminator used for stopping computation upon agent termination.

    Runtime terminator optionally executed a provided `callback` and then raises
    `SystemExit` exception in the thread performing the computation.

  Args:
    callback: callback to execute before raising exception.

  Yields:
      None.
  """
  worker_id = threading.get_ident()
  def signal_handler():
    if callback:
      callback()
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(
        ctypes.c_long(worker_id), ctypes.py_object(SystemExit))
    assert res < 2, 'Stopping worker failed'
  launchpad.register_stop_handler(signal_handler)
  try:
    yield
  except SystemExit:
    pass
  finally:
    launchpad.unregister_stop_handler(signal_handler)



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions for making TensorFlow datasets for sampling from Reverb replay."""

from typing import Optional

from acme import specs
from acme import types
from acme.adders import reverb as adders

import reverb
import tensorflow as tf


def make_reverb_dataset(
    server_address: str,
    batch_size: Optional[int] = None,
    prefetch_size: Optional[int] = None,
    table: str = adders.DEFAULT_PRIORITY_TABLE,
    num_parallel_calls: int = 12,
    max_in_flight_samples_per_worker: Optional[int] = None,
    # Deprecated kwargs.
    environment_spec: Optional[specs.EnvironmentSpec] = None,
    extra_spec: Optional[types.NestedSpec] = None,
    transition_adder: bool = False,
    convert_zero_size_to_none: bool = False,
    using_deprecated_adder: bool = False,
    sequence_length: Optional[int] = None,
) -> tf.data.Dataset:
  """Make a TensorFlow dataset backed by a Reverb trajectory replay service."""
  if environment_spec or extra_spec:
    raise ValueError(
        'The make_reverb_dataset factory function no longer requires specs as'
        ' as they should be passed as a signature to the reverb.Table when it'
        ' is created. Consider either updating your code or falling back to the'
        ' deprecated dataset factory in acme/datasets/deprecated.')

  # These are no longer used and are only kept in the call signature for
  # backward compatibility.
  del environment_spec
  del extra_spec
  del transition_adder
  del convert_zero_size_to_none
  del using_deprecated_adder
  del sequence_length

  # This is the default that used to be set by reverb.TFClient.dataset().
  if max_in_flight_samples_per_worker is None and batch_size is None:
    max_in_flight_samples_per_worker = 100
  elif max_in_flight_samples_per_worker is None:
    max_in_flight_samples_per_worker = 2 * batch_size

  def _make_dataset(unused_idx: tf.Tensor) -> tf.data.Dataset:
    dataset = reverb.TrajectoryDataset.from_table_signature(
        server_address=server_address,
        table=table,
        max_in_flight_samples_per_worker=max_in_flight_samples_per_worker)

    # Finish the pipeline: batch and prefetch.
    if batch_size:
      dataset = dataset.batch(batch_size, drop_remainder=True)

    return dataset

  # Create the dataset.
  dataset = tf.data.Dataset.range(num_parallel_calls)
  dataset = dataset.interleave(
      map_func=_make_dataset,
      cycle_length=num_parallel_calls,
      num_parallel_calls=num_parallel_calls,
      deterministic=False)

  if prefetch_size:
    dataset = dataset.prefetch(prefetch_size)

  return dataset


# TODO(b/152732834): remove this and prefer datasets.make_reverb_dataset.
make_dataset = make_reverb_dataset



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DQN agent implementation."""

from typing import Sequence

from acme import specs
from acme.agents import agent
from acme.agents import replay
from acme.agents.jax import actor_core as actor_core_lib
from acme.agents.jax import actors
from acme.agents.jax.dqn import config as dqn_config
from acme.agents.jax.dqn import learning_lib
from acme.agents.jax.dqn import losses
from acme.jax import networks as networks_lib
from acme.jax import variable_utils
import jax
import jax.numpy as jnp
import optax
import rlax


class DQNFromConfig(agent.Agent):
  """DQN agent.

  This implements a single-process DQN agent. This is a simple Q-learning
  algorithm that inserts N-step transitions into a replay buffer, and
  periodically updates its policy by sampling these transitions using
  prioritization.
  """

  def __init__(
      self,
      environment_spec: specs.EnvironmentSpec,
      network: networks_lib.FeedForwardNetwork,
      config: dqn_config.DQNConfig,
  ):
    """Initialize the agent."""
    # Data is communicated via reverb replay.
    reverb_replay = replay.make_reverb_prioritized_nstep_replay(
        environment_spec=environment_spec,
        n_step=config.n_step,
        batch_size=config.batch_size,
        max_replay_size=config.max_replay_size,
        min_replay_size=config.min_replay_size,
        priority_exponent=config.priority_exponent,
        discount=config.discount,
    )
    self._server = reverb_replay.server

    optimizer = optax.chain(
        optax.clip_by_global_norm(config.max_gradient_norm),
        optax.adam(config.learning_rate),
    )
    key_learner, key_actor = jax.random.split(jax.random.PRNGKey(config.seed))
    # The learner updates the parameters (and initializes them).
    loss_fn = losses.PrioritizedDoubleQLearning(
        discount=config.discount,
        importance_sampling_exponent=config.importance_sampling_exponent,
    )
    learner = learning_lib.SGDLearner(
        network=network,
        loss_fn=loss_fn,
        data_iterator=reverb_replay.data_iterator,
        optimizer=optimizer,
        target_update_period=config.target_update_period,
        random_key=key_learner,
        replay_client=reverb_replay.client,
    )

    # The actor selects actions according to the policy.
    assert config.epsilon is not Sequence
    def policy(params: networks_lib.Params, key: jnp.ndarray,
               observation: jnp.ndarray) -> jnp.ndarray:
      action_values = network.apply(params, observation)
      return rlax.epsilon_greedy(config.epsilon).sample(key, action_values)
    actor_core = actor_core_lib.batched_feed_forward_to_actor_core(policy)
    variable_client = variable_utils.VariableClient(learner, '')
    actor = actors.GenericActor(
        actor_core, key_actor, variable_client, reverb_replay.adder)

    super().__init__(
        actor=actor,
        learner=learner,
        min_observations=max(config.batch_size, config.min_replay_size),
        observations_per_step=config.batch_size / config.samples_per_insert,
    )


class DQN(DQNFromConfig):
  """DQN agent.

  We are in the process of migrating towards a more modular agent configuration.
  This is maintained now for compatibility.
  """

  def __init__(
      self,
      environment_spec: specs.EnvironmentSpec,
      network: networks_lib.FeedForwardNetwork,
      batch_size: int = 256,
      prefetch_size: int = 4,
      target_update_period: int = 100,
      samples_per_insert: float = 0.5,
      min_replay_size: int = 1000,
      max_replay_size: int = 1000000,
      importance_sampling_exponent: float = 0.2,
      priority_exponent: float = 0.6,
      n_step: int = 5,
      epsilon: float = 0.05,
      learning_rate: float = 1e-3,
      discount: float = 0.99,
      seed: int = 1,
  ):
    config = dqn_config.DQNConfig(
        batch_size=batch_size,
        prefetch_size=prefetch_size,
        target_update_period=target_update_period,
        samples_per_insert=samples_per_insert,
        min_replay_size=min_replay_size,
        max_replay_size=max_replay_size,
        importance_sampling_exponent=importance_sampling_exponent,
        priority_exponent=priority_exponent,
        n_step=n_step,
        epsilon=epsilon,
        learning_rate=learning_rate,
        discount=discount,
        seed=seed,
    )
    super().__init__(
        environment_spec=environment_spec,
        network=network,
        config=config,
    )



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for DQN agent."""

from absl.testing import absltest
import acme
from acme import specs
from acme.agents.jax import dqn
from acme.jax import networks as networks_lib
from acme.jax import utils
from acme.testing import fakes
import haiku as hk
import numpy as np


class DQNTest(absltest.TestCase):

  def test_dqn(self):
    # Create a fake environment to test with.
    environment = fakes.DiscreteEnvironment(
        num_actions=5,
        num_observations=10,
        obs_shape=(10, 5),
        obs_dtype=np.float32,
        episode_length=10)
    spec = specs.make_environment_spec(environment)

    def network(x):
      model = hk.Sequential([
          hk.Flatten(),
          hk.nets.MLP([50, 50, spec.actions.num_values])
      ])
      return model(x)

    # Make network purely functional
    network_hk = hk.without_apply_rng(hk.transform(network))
    dummy_obs = utils.add_batch_dim(utils.zeros_like(spec.observations))

    network = networks_lib.FeedForwardNetwork(
        init=lambda rng: network_hk.init(rng, dummy_obs),
        apply=network_hk.apply)

    # Construct the agent.
    agent = dqn.DQN(
        environment_spec=spec,
        network=network,
        batch_size=10,
        samples_per_insert=2,
        min_replay_size=10)

    # Try running the environment loop. We have no assertions here because all
    # we care about is that the agent runs without raising any errors.
    loop = acme.EnvironmentLoop(environment, agent)
    loop.run(num_episodes=20)


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the SAC agent."""

from absl.testing import absltest
import acme
from acme import specs
from acme.agents.jax.sac import agents
from acme.agents.jax.sac import config as sac_config
from acme.agents.jax.sac import networks
from acme.testing import fakes
from acme.utils import counting


class SACTest(absltest.TestCase):

  def test_sac(self):
    # Create a fake environment to test with.
    environment = fakes.ContinuousEnvironment(
        episode_length=10, action_dim=3, observation_dim=5, bounded=True)
    spec = specs.make_environment_spec(environment)

    # Create the networks.
    network = networks.make_networks(spec)

    batch_size = 10
    config = sac_config.SACConfig(
        batch_size=batch_size,
        target_entropy=sac_config.target_entropy_from_env_spec(spec),
        min_replay_size=1)
    counter = counting.Counter()
    agent = agents.SAC(
        spec=spec, network=network, config=config, seed=0, normalize_input=True,
        counter=counter)

    # Try running the environment loop. We have no assertions here because all
    # we care about is that the agent runs without raising any errors.
    loop = acme.EnvironmentLoop(environment, agent, counter=counter)
    loop.run(num_episodes=2)


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines distributed and local DAC agents, using JAX.

https://arxiv.org/pdf/1809.02925.pdf
NOTE: The original DAC implementation uses TRPO but we use TD3 here.
"""

import dataclasses
import functools
from typing import Callable

from acme import specs
from acme.agents.jax import td3
from acme.agents.jax.ail import agents
from acme.agents.jax.ail import config as ail_config
from acme.agents.jax.ail import losses
from acme.agents.jax.ail import networks as ail_networks
from acme.jax import utils
from acme.utils import loggers
import dm_env


@dataclasses.dataclass
class DACConfig:
  """Configuration options specific to DAC.

  Attributes:
    ail_config: AIL config.
    td3_config: TD3 config.
    entropy_coefficient: Entropy coefficient of the discriminator loss.
    gradient_penalty_coefficient: Coefficient for the gradient penalty term in
      the discriminator loss.
  """
  ail_config: ail_config.AILConfig
  td3_config: td3.TD3Config
  entropy_coefficient: float = 1e-3
  gradient_penalty_coefficient: float = 10.


class DistributedDAC(agents.DistributedAIL):
  """Distributed program definition for DAC."""

  def __init__(self,
               environment_factory: Callable[[bool], dm_env.Environment],
               config: DACConfig,
               *args, **kwargs):
    logger_fn = functools.partial(
        loggers.make_default_logger,
        'direct_learner',
        kwargs['log_to_bigtable'],
        time_delta=kwargs['log_every'],
        asynchronous=True,
        serialize_fn=utils.fetch_devicearray,
        steps_key='learner_steps')
    td3_agent = td3.TD3Builder(config.td3_config, logger_fn=logger_fn)

    dac_loss = losses.add_gradient_penalty(
        losses.gail_loss(entropy_coefficient=config.entropy_coefficient),
        gradient_penalty_coefficient=config.gradient_penalty_coefficient,
        gradient_penalty_target=1.)
    kwargs['discriminator_loss'] = dac_loss
    super().__init__(environment_factory, td3_agent, config.ail_config, *args,
                     **kwargs)


class DAC(agents.AIL):
  """Local agent for DAC."""

  def __init__(self, spec: specs.EnvironmentSpec,
               network: ail_networks.AILNetworks, config: DACConfig, *args,
               **kwargs):
    td3_agent = td3.TD3Builder(config.td3_config)

    dac_loss = losses.add_gradient_penalty(
        losses.gail_loss(entropy_coefficient=config.entropy_coefficient),
        gradient_penalty_coefficient=config.gradient_penalty_coefficient,
        gradient_penalty_target=1.)
    kwargs['discriminator_loss'] = dac_loss
    super().__init__(spec, td3_agent, network, config.ail_config, *args,
                     **kwargs)



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tensor framework-agnostic utilities for manipulating nested structures."""

from typing import Sequence, List, TypeVar, Any

import numpy as np
import tree

ElementType = TypeVar('ElementType')


def fast_map_structure(func, *structure):
  """Faster map_structure implementation which skips some error checking."""
  flat_structure = (tree.flatten(s) for s in structure)
  entries = zip(*flat_structure)
  # Arbitrarily choose one of the structures of the original sequence (the last)
  # to match the structure for the flattened sequence.
  return tree.unflatten_as(structure[-1], [func(*x) for x in entries])


def fast_map_structure_with_path(func, *structure):
  """Faster map_structure_with_path implementation."""
  head_entries_with_path = tree.flatten_with_path(structure[0])
  if len(structure) > 1:
    tail_entries = (tree.flatten(s) for s in structure[1:])
    entries_with_path = [
        e[0] + e[1:] for e in zip(head_entries_with_path, *tail_entries)
    ]
  else:
    entries_with_path = head_entries_with_path
  # Arbitrarily choose one of the structures of the original sequence (the last)
  # to match the structure for the flattened sequence.
  return tree.unflatten_as(structure[-1], [func(*x) for x in entries_with_path])


def stack_sequence_fields(sequence: Sequence[ElementType]) -> ElementType:
  """Stacks a list of identically nested objects.

  This takes a sequence of identically nested objects and returns a single
  nested object whose ith leaf is a stacked numpy array of the corresponding
  ith leaf from each element of the sequence.

  For example, if `sequence` is:

  ```python
  [{
        'action': np.array([1.0]),
        'observation': (np.array([0.0, 1.0, 2.0]),),
        'reward': 1.0
   }, {
        'action': np.array([0.5]),
        'observation': (np.array([1.0, 2.0, 3.0]),),
        'reward': 0.0
   }, {
        'action': np.array([0.3]),1
        'observation': (np.array([2.0, 3.0, 4.0]),),
        'reward': 0.5
   }]
  ```

  Then this function will return:

  ```python
  {
      'action': np.array([....])         # array shape = [3 x 1]
      'observation': (np.array([...]),)  # array shape = [3 x 3]
      'reward': np.array([...])          # array shape = [3]
  }
  ```

  Note that the 'observation' entry in the above example has two levels of
  nesting, i.e it is a tuple of arrays.

  Args:
    sequence: a list of identically nested objects.

  Returns:
    A nested object with numpy.

  Raises:
    ValueError: If `sequence` is an empty sequence.
  """
  # Handle empty input sequences.
  if not sequence:
    raise ValueError('Input sequence must not be empty')

  # Default to asarray when arrays don't have the same shape to be compatible
  # with old behaviour.
  try:
    return fast_map_structure(lambda *values: np.stack(values), *sequence)
  except ValueError:
    return fast_map_structure(lambda *values: np.asarray(values), *sequence)


def unstack_sequence_fields(struct: ElementType,
                            batch_size: int) -> List[ElementType]:
  """Converts a struct of batched arrays to a list of structs.

  This is effectively the inverse of `stack_sequence_fields`.

  Args:
    struct: An (arbitrarily nested) structure of arrays.
    batch_size: The length of the leading dimension of each array in the struct.
      This is assumed to be static and known.

  Returns:
    A list of structs with the same structure as `struct`, where each leaf node
     is an unbatched element of the original leaf node.
  """

  return [
      tree.map_structure(lambda s, i=i: s[i], struct) for i in range(batch_size)
  ]


def broadcast_structures(*args: Any) -> Any:
  """Returns versions of the arguments that give them the same nested structure.

  Any nested items in *args must have the same structure.

  Any non-nested item will be replaced with a nested version that shares that
  structure. The leaves will all be references to the same original non-nested
  item.

  If all *args are nested, or all *args are non-nested, this function will
  return *args unchanged.

  Example:
  ```
  a = ('a', 'b')
  b = 'c'
  tree_a, tree_b = broadcast_structure(a, b)
  tree_a
  > ('a', 'b')
  tree_b
  > ('c', 'c')
  ```

  Args:
    *args: A Sequence of nested or non-nested items.

  Returns:
    `*args`, except with all items sharing the same nest structure.
  """
  if not args:
    return

  reference_tree = None
  for arg in args:
    if tree.is_nested(arg):
      reference_tree = arg
      break

  # If reference_tree is None then none of args are nested and we can skip over
  # the rest of this function, which would be a no-op.
  if reference_tree is None:
    return args

  def mirror_structure(value, reference_tree):
    if tree.is_nested(value):
      # Use check_types=True so that the types of the trees we construct aren't
      # dependent on our arbitrary choice of which nested arg to use as the
      # reference_tree.
      tree.assert_same_structure(value, reference_tree, check_types=True)
      return value
    else:
      return tree.map_structure(lambda _: value, reference_tree)

  return tuple(mirror_structure(arg, reference_tree) for arg in args)


def tree_map(f):
  """Transforms `f` into a tree-mapped version."""

  def mapped_f(*structures):
    return tree.map_structure(f, *structures)

  return mapped_f



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines distributed and local PPO agents, using JAX."""

import functools
from typing import Callable, Optional

from acme import specs
from acme.agents.jax import normalization
from acme.agents.jax.ppo import builder
from acme.agents.jax.ppo import config as ppo_config
from acme.agents.jax.ppo import networks as ppo_networks
from acme.jax import utils
from acme.jax.layouts import distributed_layout
from acme.jax.layouts import local_layout
from acme.utils import counting
from acme.utils import loggers
import dm_env


NetworkFactory = Callable[[specs.EnvironmentSpec], ppo_networks.PPONetworks]


class DistributedPPO(distributed_layout.DistributedLayout):
  """Distributed program definition for PPO."""

  def __init__(
      self,
      environment_factory: Callable[[bool], dm_env.Environment],
      network_factory: NetworkFactory,
      config: ppo_config.PPOConfig,
      seed: int,
      num_actors: int,
      normalize_input: bool = False,
      logger_fn: Optional[Callable[[], loggers.Logger]] = None,
      save_reverb_logs: bool = False,
      log_every: float = 10.0,
      max_number_of_steps: Optional[int] = None,
  ):
    logger_fn = logger_fn or functools.partial(
        loggers.make_default_logger,
        'learner',
        save_reverb_logs,
        time_delta=log_every,
        asynchronous=True,
        serialize_fn=utils.fetch_devicearray,
        steps_key='learner_steps')
    ppo_builder = builder.PPOBuilder(config, logger_fn=logger_fn)
    if normalize_input:
      environment_spec = specs.make_environment_spec(environment_factory(False))
      # Two batch dimensions: [num_sequences, num_steps, ...]
      batch_dims = (0, 1)
      ppo_builder = normalization.NormalizationBuilder(
          ppo_builder,
          environment_spec,
          is_sequence_based=True,
          batch_dims=batch_dims)
    eval_policy_factory = (
        lambda networks: ppo_networks.make_inference_fn(networks, True))
    super().__init__(
        seed=seed,
        environment_factory=lambda: environment_factory(False),
        network_factory=network_factory,
        builder=ppo_builder,
        policy_network=ppo_networks.make_inference_fn,
        evaluator_factories=[
            distributed_layout.default_evaluator(
                environment_factory=lambda: environment_factory(True),
                network_factory=network_factory,
                builder=ppo_builder,
                policy_factory=eval_policy_factory,
                log_to_bigtable=save_reverb_logs)
        ],
        num_actors=num_actors,
        prefetch_size=config.prefetch_size,
        max_number_of_steps=max_number_of_steps,
        log_to_bigtable=save_reverb_logs,
        actor_logger_fn=distributed_layout.get_default_logger_fn(
            save_reverb_logs, log_every),
    )


class PPO(local_layout.LocalLayout):
  """Local agent for PPO."""

  def __init__(
      self,
      spec: specs.EnvironmentSpec,
      networks: ppo_networks.PPONetworks,
      config: ppo_config.PPOConfig,
      seed: int,
      workdir: Optional[str] = '~/acme',
      normalize_input: bool = False,
      counter: Optional[counting.Counter] = None,
      logger_fn: Optional[Callable[[], loggers.Logger]] = None,
  ):
    logger_fn = logger_fn or (lambda: None)
    ppo_builder = builder.PPOBuilder(config, logger_fn=logger_fn)
    if normalize_input:
      # Two batch dimensions: [num_sequences, num_steps, ...]
      batch_dims = (0, 1)
      ppo_builder = normalization.NormalizationBuilder(
          ppo_builder, spec, is_sequence_based=True, batch_dims=batch_dims)
    self.builder = ppo_builder
    super().__init__(
        seed=seed,
        environment_spec=spec,
        builder=ppo_builder,
        networks=networks,
        policy_network=ppo_networks.make_inference_fn(networks),
        batch_size=config.batch_size,
        # TODO(sinopalnikov): move it to the experiment config
        workdir=workdir,
        counter=counter,
    )



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Shared helpers for different experiment flavours."""

from typing import Mapping, Sequence

from acme import specs
from acme import types
from acme.tf import networks
from acme.tf import utils as tf2_utils

import numpy as np
import sonnet as snt


def make_default_networks(
    action_spec: specs.BoundedArray,
    policy_layer_sizes: Sequence[int] = (256, 256, 256),
    critic_layer_sizes: Sequence[int] = (512, 512, 256),
    vmin: float = -150.,
    vmax: float = 150.,
    num_atoms: int = 51,
) -> Mapping[str, types.TensorTransformation]:
  """Creates networks used by the agent."""

  # Get total number of action dimensions from action spec.
  num_dimensions = np.prod(action_spec.shape, dtype=int)

  # Create the shared observation network; here simply a state-less operation.
  observation_network = tf2_utils.batch_concat

  # Create the policy network.
  policy_network = snt.Sequential([
      networks.LayerNormMLP(policy_layer_sizes, activate_final=True),
      networks.NearZeroInitializedLinear(num_dimensions),
      networks.TanhToSpec(action_spec),
  ])

  # Create the critic network.
  critic_network = snt.Sequential([
      # The multiplexer concatenates the observations/actions.
      networks.CriticMultiplexer(),
      networks.LayerNormMLP(critic_layer_sizes, activate_final=True),
      networks.DiscreteValuedHead(vmin, vmax, num_atoms),
  ])

  return {
      'policy': policy_network,
      'critic': critic_network,
      'observation': observation_network,
  }



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""PPO network definitions."""

import dataclasses
from typing import Any, Callable, Optional, Sequence

from acme import specs
from acme.agents.jax import actor_core as actor_core_lib
from acme.jax import networks as networks_lib
from acme.jax import utils
import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np


EntropyFn = Callable[[Any], jnp.ndarray]


@dataclasses.dataclass
class PPONetworks:
  """Network and pure functions for the PPO agent.

  If 'network' returns tfd.Distribution, you can use make_ppo_networks() to
  create this object properly.
  If one is building this object manually, one has a freedom to make 'network'
  object return anything that is later being passed as input to
  log_prob/entropy/sample functions to perform the corresponding computations.
  """
  network: networks_lib.FeedForwardNetwork
  log_prob: networks_lib.LogProbFn
  entropy: EntropyFn
  sample: networks_lib.SampleFn
  sample_eval: Optional[networks_lib.SampleFn] = None


def make_inference_fn(ppo_networks: PPONetworks, evaluation: bool = False
                      ) -> actor_core_lib.FeedForwardPolicyWithExtra:
  """Returns a function to be used for inference by a PPO actor."""

  def inference(params: networks_lib.Params, key: networks_lib.PRNGKey,
                observations: networks_lib.Observation):
    distribution, _ = ppo_networks.network.apply(params, observations)
    if evaluation and ppo_networks.sample_eval:
      actions = ppo_networks.sample_eval(distribution, key)
    else:
      actions = ppo_networks.sample(distribution, key)
    if evaluation:
      return actions, {}
    log_prob = ppo_networks.log_prob(distribution, actions)
    return actions, {'log_prob': log_prob}
  return inference


def make_ppo_networks(network: networks_lib.FeedForwardNetwork) -> PPONetworks:
  """Constructs a PPONetworks instance from the given FeedForwardNetwork.

  Args:
    network: a transformed Haiku network that takes in observations and returns
      the action distribution and value.

  Returns:
    A PPONetworks instance with pure functions wrapping the input network.
  """
  return PPONetworks(
      network=network,
      log_prob=lambda distribution, action: distribution.log_prob(action),
      entropy=lambda distribution: distribution.entropy(),
      sample=lambda distribution, key: distribution.sample(seed=key),
      sample_eval=lambda distribution, key: distribution.mode())


def make_atari_networks(
    environment_spec: specs.EnvironmentSpec,
    hidden_layer_sizes: Sequence[int] = (512,),
) -> PPONetworks:
  """Creates networks used by the agent for Atari environments."""

  num_actions = environment_spec.actions.num_values

  def forward_fn(inputs):
    policy_value_network = hk.Sequential([
        networks_lib.AtariTorso(),
        hk.nets.MLP(hidden_layer_sizes, activation=jax.nn.relu),
        networks_lib.CategoricalValueHead(num_values=num_actions)
    ])
    return policy_value_network(inputs)

  forward_fn = hk.without_apply_rng(hk.transform(forward_fn))
  dummy_obs = utils.zeros_like(environment_spec.observations)
  dummy_obs = utils.add_batch_dim(dummy_obs)  # Dummy 'sequence' dim.
  network = networks_lib.FeedForwardNetwork(
      lambda rng: forward_fn.init(rng, dummy_obs), forward_fn.apply)
  # Create PPONetworks to add functionality required by the agent.
  return make_ppo_networks(network)


def make_gym_networks(
    environment_spec: specs.EnvironmentSpec,
    policy_layer_sizes: Sequence[int] = (64, 64),
    value_layer_sizes: Sequence[int] = (64, 64),
) -> PPONetworks:
  """Creates networks to be used by the agent for OpenAI Gym environments."""

  # Get total number of action dimensions from action spec.
  num_dimensions = np.prod(environment_spec.actions.shape, dtype=int)

  def forward_fn(inputs):
    policy_network = hk.Sequential([
        utils.batch_concat,
        hk.nets.MLP(policy_layer_sizes, activation=jnp.tanh),
        # Note: we don't respect bounded action specs here and instead
        # rely on CanonicalSpecWrapper to clip actions accordingly.
        networks_lib.MultivariateNormalDiagHead(num_dimensions)
    ])
    value_network = hk.Sequential([
        utils.batch_concat,
        hk.nets.MLP(value_layer_sizes, activation=jnp.tanh),
        hk.Linear(1), lambda x: jnp.squeeze(x, axis=-1)
    ])

    action_distribution = policy_network(inputs)
    value = value_network(inputs)
    return (action_distribution, value)

  # Transform into pure functions.
  forward_fn = hk.without_apply_rng(hk.transform(forward_fn))

  dummy_obs = utils.zeros_like(environment_spec.observations)
  dummy_obs = utils.add_batch_dim(dummy_obs)  # Dummy 'sequence' dim.
  network = networks_lib.FeedForwardNetwork(
      lambda rng: forward_fn.init(rng, dummy_obs), forward_fn.apply)
  # Create PPONetworks to add functionality required by the agent.
  return make_ppo_networks(network)



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An OpenSpiel multi-agent/environment training loop."""

import operator
import time
from typing import Optional, Sequence

from acme import core
from acme.utils import counting
from acme.utils import loggers
from acme.wrappers import open_spiel_wrapper
import dm_env
from dm_env import specs
import numpy as np
import tree

# pytype: disable=import-error
import pyspiel
# pytype: enable=import-error


class OpenSpielEnvironmentLoop(core.Worker):
  """An OpenSpiel RL environment loop.

  This takes `Environment` and list of `Actor` instances and coordinates their
  interaction. Agents are updated if `should_update=True`. This can be used as:

    loop = EnvironmentLoop(environment, actors)
    loop.run(num_episodes)

  A `Counter` instance can optionally be given in order to maintain counts
  between different Acme components. If not given a local Counter will be
  created to maintain counts between calls to the `run` method.

  A `Logger` instance can also be passed in order to control the output of the
  loop. If not given a platform-specific default logger will be used as defined
  by utils.loggers.make_default_logger. A string `label` can be passed to easily
  change the label associated with the default logger; this is ignored if a
  `Logger` instance is given.
  """

  def __init__(
      self,
      environment: open_spiel_wrapper.OpenSpielWrapper,
      actors: Sequence[core.Actor],
      counter: Optional[counting.Counter] = None,
      logger: Optional[loggers.Logger] = None,
      should_update: bool = True,
      label: str = 'open_spiel_environment_loop',
  ):
    # Internalize agent and environment.
    self._environment = environment
    self._actors = actors
    self._counter = counter or counting.Counter()
    self._logger = logger or loggers.make_default_logger(label)
    self._should_update = should_update

    # Track information necessary to coordinate updates among multiple actors.
    self._observed_first = [False] * len(self._actors)
    self._prev_actions = [pyspiel.INVALID_ACTION] * len(self._actors)

  def _send_observation(self, timestep: dm_env.TimeStep, player: int):
    # If terminal all actors must update
    if player == pyspiel.PlayerId.TERMINAL:
      for player_id in range(len(self._actors)):
        # Note: we must account for situations where the first observation
        # is a terminal state, e.g. if an opponent folds in poker before we get
        # to act.
        if self._observed_first[player_id]:
          player_timestep = self._get_player_timestep(timestep, player_id)
          self._actors[player_id].observe(self._prev_actions[player_id],
                                          player_timestep)
          if self._should_update:
            self._actors[player_id].update()
      self._observed_first = [False] * len(self._actors)
      self._prev_actions = [pyspiel.INVALID_ACTION] * len(self._actors)
    else:
      if not self._observed_first[player]:
        player_timestep = dm_env.TimeStep(
            observation=timestep.observation[player],
            reward=None,
            discount=None,
            step_type=dm_env.StepType.FIRST)
        self._actors[player].observe_first(player_timestep)
        self._observed_first[player] = True
      else:
        player_timestep = self._get_player_timestep(timestep, player)
        self._actors[player].observe(self._prev_actions[player],
                                     player_timestep)
        if self._should_update:
          self._actors[player].update()

  def _get_action(self, timestep: dm_env.TimeStep, player: int) -> int:
    self._prev_actions[player] = self._actors[player].select_action(
        timestep.observation[player])
    return self._prev_actions[player]

  def _get_player_timestep(self, timestep: dm_env.TimeStep,
                           player: int) -> dm_env.TimeStep:
    return dm_env.TimeStep(observation=timestep.observation[player],
                           reward=timestep.reward[player],
                           discount=timestep.discount[player],
                           step_type=timestep.step_type)

  def run_episode(self) -> loggers.LoggingData:
    """Run one episode.

    Each episode is a loop which interacts first with the environment to get an
    observation and then give that observation to the agent in order to retrieve
    an action.

    Returns:
      An instance of `loggers.LoggingData`.
    """
    # Reset any counts and start the environment.
    start_time = time.time()
    episode_steps = 0

    # For evaluation, this keeps track of the total undiscounted reward
    # for each player accumulated during the episode.
    multiplayer_reward_spec = specs.BoundedArray(
        (self._environment.game.num_players(),),
        np.float32,
        minimum=self._environment.game.min_utility(),
        maximum=self._environment.game.max_utility())
    episode_return = tree.map_structure(_generate_zeros_from_spec,
                                        multiplayer_reward_spec)

    timestep = self._environment.reset()

    # Make the first observation.
    self._send_observation(timestep, self._environment.current_player)

    # Run an episode.
    while not timestep.last():
      # Generate an action from the agent's policy and step the environment.
      if self._environment.is_turn_based:
        action_list = [
            self._get_action(timestep, self._environment.current_player)
        ]
      else:
        # FIXME: Support simultaneous move games.
        raise ValueError('Currently only supports sequential games.')

      timestep = self._environment.step(action_list)

      # Have the agent observe the timestep and let the actor update itself.
      self._send_observation(timestep, self._environment.current_player)

      # Book-keeping.
      episode_steps += 1

      # Equivalent to: episode_return += timestep.reward
      # We capture the return value because if timestep.reward is a JAX
      # DeviceArray, episode_return will not be mutated in-place. (In all other
      # cases, the returned episode_return will be the same object as the
      # argument episode_return.)
      episode_return = tree.map_structure(operator.iadd,
                                          episode_return,
                                          timestep.reward)

    # Record counts.
    counts = self._counter.increment(episodes=1, steps=episode_steps)

    # Collect the results and combine with counts.
    steps_per_second = episode_steps / (time.time() - start_time)
    result = {
        'episode_length': episode_steps,
        'episode_return': episode_return,
        'steps_per_second': steps_per_second,
    }
    result.update(counts)
    return result

  def run(self,
          num_episodes: Optional[int] = None,
          num_steps: Optional[int] = None):
    """Perform the run loop.

    Run the environment loop either for `num_episodes` episodes or for at
    least `num_steps` steps (the last episode is always run until completion,
    so the total number of steps may be slightly more than `num_steps`).
    At least one of these two arguments has to be None.

    Upon termination of an episode a new episode will be started. If the number
    of episodes and the number of steps are not given then this will interact
    with the environment infinitely.

    Args:
      num_episodes: number of episodes to run the loop for.
      num_steps: minimal number of steps to run the loop for.

    Raises:
      ValueError: If both 'num_episodes' and 'num_steps' are not None.
    """

    if not (num_episodes is None or num_steps is None):
      raise ValueError('Either "num_episodes" or "num_steps" should be None.')

    def should_terminate(episode_count: int, step_count: int) -> bool:
      return ((num_episodes is not None and episode_count >= num_episodes) or
              (num_steps is not None and step_count >= num_steps))

    episode_count, step_count = 0, 0
    while not should_terminate(episode_count, step_count):
      result = self.run_episode()
      episode_count += 1
      step_count += result['episode_length']
      # Log the given results.
      self._logger.write(result)


def _generate_zeros_from_spec(spec: specs.Array) -> np.ndarray:
  return np.zeros(spec.shape, spec.dtype)



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ValueDice networks definition."""


import dataclasses
from typing import Callable, Optional, Tuple

from acme import specs
from acme.agents.jax import actor_core as actor_core_lib
from acme.jax import networks as networks_lib
from acme.jax import utils
import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np


@dataclasses.dataclass
class ValueDiceNetworks:
  """ValueDice networks."""
  policy_network: networks_lib.FeedForwardNetwork
  nu_network: networks_lib.FeedForwardNetwork
  # Functions for actors and evaluators, resp., to sample actions.
  sample: networks_lib.SampleFn
  sample_eval: Optional[networks_lib.SampleFn] = None
  # Function that transforms an action before a mixture is applied, typically
  # the identity for continuous actions and one-hot encoding for discrete
  # actions.
  encode_action: Callable[[networks_lib.Action], jnp.ndarray] = lambda x: x


def apply_policy_and_sample(
    networks: ValueDiceNetworks,
    eval_mode: bool = False) -> actor_core_lib.FeedForwardPolicy:
  """Returns a function that computes actions."""
  sample_fn = networks.sample if not eval_mode else networks.sample_eval
  if not sample_fn:
    raise ValueError('sample function is not provided')

  def apply_and_sample(params, key, obs):
    return sample_fn(networks.policy_network.apply(params, obs), key)
  return apply_and_sample


def make_networks(
    spec: specs.EnvironmentSpec,
    hidden_layer_sizes: Tuple[int, ...] = (256, 256)) -> ValueDiceNetworks:
  """Creates networks used by the agent."""

  num_dimensions = np.prod(spec.actions.shape, dtype=int)

  def _actor_fn(obs):
    network = hk.Sequential([
        hk.nets.MLP(
            list(hidden_layer_sizes),
            w_init=hk.initializers.VarianceScaling(1.0, 'fan_in', 'uniform'),
            activation=jax.nn.relu,
            activate_final=True),
        networks_lib.NormalTanhDistribution(num_dimensions),
    ])
    return network(obs)

  def _nu_fn(obs, action):
    network = hk.Sequential([
        hk.nets.MLP(
            list(hidden_layer_sizes) + [1],
            w_init=hk.initializers.VarianceScaling(1.0, 'fan_in', 'uniform'),
            activation=jax.nn.relu),
    ])
    return network(jnp.concatenate([obs, action], axis=-1))

  policy = hk.without_apply_rng(hk.transform(_actor_fn))
  nu = hk.without_apply_rng(hk.transform(_nu_fn))

  # Create dummy observations and actions to create network parameters.
  dummy_action = utils.zeros_like(spec.actions)
  dummy_obs = utils.zeros_like(spec.observations)
  dummy_action = utils.add_batch_dim(dummy_action)
  dummy_obs = utils.add_batch_dim(dummy_obs)

  return ValueDiceNetworks(
      policy_network=networks_lib.FeedForwardNetwork(
          lambda key: policy.init(key, dummy_obs), policy.apply),
      nu_network=networks_lib.FeedForwardNetwork(
          lambda key: nu.init(key, dummy_obs, dummy_action), nu.apply),
      sample=lambda params, key: params.sample(seed=key),
      sample_eval=lambda params, key: params.mode())



# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the IMPALA loss function."""

from absl.testing import absltest
from acme.adders import reverb as adders
from acme.jax.losses import impala
from acme.utils.tree_utils import tree_map

import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
import reverb


class ImpalaTest(absltest.TestCase):

  def test_shapes(self):

    #
    batch_size = 2
    sequence_len = 3
    num_actions = 5
    hidden_size = 7

    # Define a trivial recurrent actor-critic network.
    @hk.without_apply_rng
    @hk.transform
    def unroll_fn_transformed(observations, state):
      lstm = hk.LSTM(hidden_size)
      embedding, state = hk.dynamic_unroll(lstm, observations, state)
      logits = hk.Linear(num_actions)(embedding)
      values = jnp.squeeze(hk.Linear(1)(embedding), axis=-1)

      return (logits, values), state

    @hk.without_apply_rng
    @hk.transform
    def initial_state_fn():
      return hk.LSTM(hidden_size).initial_state(None)

    # Initial recurrent network state.
    initial_state = initial_state_fn.apply(None)

    # Make some fake data.
    observations = np.ones(shape=(sequence_len, 50))
    actions = np.random.randint(num_actions, size=sequence_len)
    rewards = np.random.rand(sequence_len)
    discounts = np.ones(shape=(sequence_len,))

    batch_tile = tree_map(lambda x: np.tile(x, [batch_size, *([1] * x.ndim)]))
    seq_tile = tree_map(lambda x: np.tile(x, [sequence_len, *([1] * x.ndim)]))

    extras = {
        'logits': np.random.rand(sequence_len, num_actions),
        'core_state': seq_tile(initial_state),
    }

    # Package up the data into a ReverbSample.
    data = adders.Step(
        observations,
        actions,
        rewards,
        discounts,
        extras=extras,
        start_of_episode=())
    data = batch_tile(data)
    sample = reverb.ReplaySample(info=None, data=data)

    # Initialise parameters.
    rng = hk.PRNGSequence(1)
    params = unroll_fn_transformed.init(next(rng), observations, initial_state)

    # Make loss function.
    loss_fn = impala.impala_loss(
        unroll_fn_transformed.apply, discount=0.99)

    # Return value should be scalar.
    loss = loss_fn(params, sample)
    loss = jax.device_get(loss)
    self.assertEqual(loss.shape, ())


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tree_utils."""

import functools
from typing import Sequence

from absl.testing import absltest
from acme.utils import tree_utils
import numpy as np
import tree

TEST_SEQUENCE = [
    {
        'action': np.array([1.0]),
        'observation': (np.array([0.0, 1.0, 2.0]),),
        'reward': np.array(1.0),
    },
    {
        'action': np.array([0.5]),
        'observation': (np.array([1.0, 2.0, 3.0]),),
        'reward': np.array(0.0),
    },
    {
        'action': np.array([0.3]),
        'observation': (np.array([2.0, 3.0, 4.0]),),
        'reward': np.array(0.5),
    },
]


class SequenceStackTest(absltest.TestCase):
  """Tests for various tree utilities."""

  def test_stack_sequence_fields(self):
    """Tests that `stack_sequence_fields` behaves correctly on nested data."""

    stacked = tree_utils.stack_sequence_fields(TEST_SEQUENCE)

    # Check that the stacked output has the correct structure.
    tree.assert_same_structure(stacked, TEST_SEQUENCE[0])

    # Check that the leaves have the correct array shapes.
    self.assertEqual(stacked['action'].shape, (3, 1))
    self.assertEqual(stacked['observation'][0].shape, (3, 3))
    self.assertEqual(stacked['reward'].shape, (3,))

    # Check values.
    self.assertEqual(stacked['observation'][0].tolist(), [
        [0., 1., 2.],
        [1., 2., 3.],
        [2., 3., 4.],
    ])
    self.assertEqual(stacked['action'].tolist(), [[1.], [0.5], [0.3]])
    self.assertEqual(stacked['reward'].tolist(), [1., 0., 0.5])

  def test_unstack_sequence_fields(self):
    """Tests that `unstack_sequence_fields(stack_sequence_fields(x)) == x`."""
    stacked = tree_utils.stack_sequence_fields(TEST_SEQUENCE)
    batch_size = len(TEST_SEQUENCE)
    unstacked = tree_utils.unstack_sequence_fields(stacked, batch_size)
    tree.map_structure(np.testing.assert_array_equal, unstacked, TEST_SEQUENCE)

  def test_fast_map_structure_with_path(self):
    structure = {
        'a': {
            'b': np.array([0.0])
        },
        'c': (np.array([1.0]), np.array([2.0])),
        'd': [np.array(3.0), np.array(4.0)],
    }

    def map_fn(path: Sequence[str], x: np.ndarray, y: np.ndarray):
      return x + y + len(path)

    single_arg_map_fn = functools.partial(map_fn, y=np.array([0.0]))

    expected_mapped_structure = (
        tree.map_structure_with_path(single_arg_map_fn, structure))
    mapped_structure = (
        tree_utils.fast_map_structure_with_path(single_arg_map_fn, structure))
    self.assertEqual(mapped_structure, expected_mapped_structure)

    expected_double_mapped_structure = (
        tree.map_structure_with_path(map_fn, structure, mapped_structure))
    double_mapped_structure = (
        tree_utils.fast_map_structure_with_path(map_fn, structure,
                                                mapped_structure))
    self.assertEqual(double_mapped_structure, expected_double_mapped_structure)


if __name__ == '__main__':
  absltest.main()



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The base agent interface."""

from typing import List

from acme import core
from acme import types
# Internal imports.

import dm_env
import numpy as np


def _calculate_num_learner_steps(num_observations: int,
                                 min_observations: int,
                                 observations_per_step: float) -> int:
  """Calculates the number of learner steps to do at step=num_observations."""
  n = num_observations - min_observations
  if n < 0:
    # Do not do any learner steps until you have seen min_observations.
    return 0
  if observations_per_step > 1:
    # One batch every 1/obs_per_step observations, otherwise zero.
    return int(n % int(observations_per_step) == 0)
  else:
    # Always return 1/obs_per_step batches every observation.
    return int(1 / observations_per_step)


class Agent(core.Actor, core.VariableSource):
  """Agent class which combines acting and learning.

  This provides an implementation of the `Actor` interface which acts and
  learns. It takes as input instances of both `acme.Actor` and `acme.Learner`
  classes, and implements the policy, observation, and update methods which
  defer to the underlying actor and learner.

  The only real logic implemented by this class is that it controls the number
  of observations to make before running a learner step. This is done by
  passing the number of `min_observations` to use and a ratio of
  `observations_per_step` := num_actor_actions / num_learner_steps.

  Note that the number of `observations_per_step` can also be in the range[0, 1]
  in order to allow the agent to take more than 1 learner step per action.
  """

  def __init__(self, actor: core.Actor, learner: core.Learner,
               min_observations: int, observations_per_step: float):
    self._actor = actor
    self._learner = learner
    self._min_observations = min_observations
    self._observations_per_step = observations_per_step
    self._num_observations = 0

  def select_action(self, observation: types.NestedArray) -> types.NestedArray:
    return self._actor.select_action(observation)

  def observe_first(self, timestep: dm_env.TimeStep):
    self._actor.observe_first(timestep)

  def observe(self, action: types.NestedArray, next_timestep: dm_env.TimeStep):
    self._num_observations += 1
    self._actor.observe(action, next_timestep)

  def update(self):
    num_steps = _calculate_num_learner_steps(
        num_observations=self._num_observations,
        min_observations=self._min_observations,
        observations_per_step=self._observations_per_step,
    )
    for _ in range(num_steps):
      # Run learner steps (usually means gradient steps).
      self._learner.step()
    if num_steps > 0:
      # Update the actor weights when learner updates.
      self._actor.update()

  def get_variables(self, names: List[str]) -> List[List[np.ndarray]]:
    return self._learner.get_variables(names)


# Internal class.



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Importance weighted advantage actor-critic (IMPALA) agent implementation."""

from typing import Callable, Optional, Union

import acme
from acme import specs
from acme import types
from acme.agents import replay
from acme.agents.jax.impala import acting
from acme.agents.jax.impala import config as impala_config
from acme.agents.jax.impala import learning
from acme.agents.jax.impala import networks as impala_networks
from acme.agents.jax.impala import types as impala_types
from acme.jax import networks as networks_lib
from acme.jax import variable_utils
from acme.utils import counting
from acme.utils import loggers
import dm_env
import haiku as hk
import jax
import numpy as np
import optax


class IMPALAFromConfig(acme.Actor):
  """IMPALA Agent."""

  def __init__(
      self,
      environment_spec: specs.EnvironmentSpec,
      forward_fn: impala_types.PolicyValueFn,
      unroll_init_fn: impala_types.PolicyValueInitFn,
      unroll_fn: impala_types.PolicyValueFn,
      initial_state_init_fn: impala_types.RecurrentStateInitFn,
      initial_state_fn: impala_types.RecurrentStateFn,
      config: impala_config.IMPALAConfig,
      counter: Optional[counting.Counter] = None,
      logger: Optional[loggers.Logger] = None,
  ):
    networks = impala_networks.IMPALANetworks(
        forward_fn=forward_fn,
        unroll_init_fn=unroll_init_fn,
        unroll_fn=unroll_fn,
        initial_state_init_fn=initial_state_init_fn,
        initial_state_fn=initial_state_fn,
    )

    self._config = config

    # Data is handled by the reverb replay queue.
    num_actions = environment_spec.actions.num_values
    self._logger = logger or loggers.TerminalLogger('agent')

    key, key_initial_state = jax.random.split(
        jax.random.PRNGKey(self._config.seed))
    params = initial_state_init_fn(key_initial_state)
    extra_spec = {
        'core_state': initial_state_fn(params),
        'logits': np.ones(shape=(num_actions,), dtype=np.float32)
    }

    reverb_queue = replay.make_reverb_online_queue(
        environment_spec=environment_spec,
        extra_spec=extra_spec,
        max_queue_size=self._config.max_queue_size,
        sequence_length=self._config.sequence_length,
        sequence_period=self._config.sequence_period,
        batch_size=self._config.batch_size,
    )
    self._server = reverb_queue.server
    self._can_sample = reverb_queue.can_sample

    # Make the learner.
    optimizer = optax.chain(
        optax.clip_by_global_norm(self._config.max_gradient_norm),
        optax.adam(self._config.learning_rate),
    )
    key_learner, key_actor = jax.random.split(key)
    self._learner = learning.IMPALALearner(
        networks=networks,
        iterator=reverb_queue.data_iterator,
        random_key=key_learner,
        counter=counter,
        logger=logger,
        optimizer=optimizer,
        discount=self._config.discount,
        entropy_cost=self._config.entropy_cost,
        baseline_cost=self._config.baseline_cost,
        max_abs_reward=self._config.max_abs_reward,
    )

    # Make the actor.
    variable_client = variable_utils.VariableClient(self._learner, key='policy')
    self._actor = acting.IMPALAActor(
        forward_fn=jax.jit(forward_fn, backend='cpu'),
        initial_state_init_fn=initial_state_init_fn,
        initial_state_fn=initial_state_fn,
        rng=hk.PRNGSequence(key_actor),
        adder=reverb_queue.adder,
        variable_client=variable_client,
    )

  def observe_first(self, timestep: dm_env.TimeStep):
    self._actor.observe_first(timestep)

  def observe(
      self,
      action: impala_types.Action,
      next_timestep: dm_env.TimeStep,
  ):
    self._actor.observe(action, next_timestep)

  def update(self, wait: bool = False):
    should_update_actor = False
    # Run a number of learner steps (usually gradient steps).
    while self._can_sample():
      self._learner.step()
      should_update_actor = True
    if should_update_actor:
      # Update actor weights after learner.
      self._actor.update(wait)

  def select_action(self, observation: np.ndarray) -> int:
    return self._actor.select_action(observation)


class IMPALA(IMPALAFromConfig):
  """IMPALA agent.

  We are in the process of migrating towards a more modular agent configuration.
  This is maintained now for compatibility.
  """

  def __init__(
      self,
      environment_spec: specs.EnvironmentSpec,
      forward_fn: networks_lib.PolicyValueRNN,
      unroll_fn: networks_lib.PolicyValueRNN,
      initial_state_fn: Callable[[], hk.LSTMState],
      sequence_length: int,
      sequence_period: int,
      counter: Optional[counting.Counter] = None,
      logger: Optional[loggers.Logger] = None,
      discount: float = 0.99,
      max_queue_size: Union[int, types.Batches] = types.Batches(10),
      batch_size: int = 16,
      learning_rate: float = 1e-3,
      entropy_cost: float = 0.01,
      baseline_cost: float = 0.5,
      seed: int = 0,
      max_abs_reward: float = np.inf,
      max_gradient_norm: float = np.inf,
  ):
    networks = impala_networks.make_haiku_networks(
        env_spec=environment_spec,
        forward_fn=forward_fn, initial_state_fn=initial_state_fn,
        unroll_fn=unroll_fn)

    config = impala_config.IMPALAConfig(
        sequence_length=sequence_length,
        sequence_period=sequence_period,
        discount=discount,
        max_queue_size=max_queue_size,
        batch_size=batch_size,
        learning_rate=learning_rate,
        entropy_cost=entropy_cost,
        baseline_cost=baseline_cost,
        seed=seed,
        max_abs_reward=max_abs_reward,
        max_gradient_norm=max_gradient_norm,
    )

    super().__init__(
        environment_spec=environment_spec,
        forward_fn=networks.forward_fn,
        unroll_init_fn=networks.unroll_init_fn,
        unroll_fn=networks.unroll_fn,
        initial_state_init_fn=networks.initial_state_init_fn,
        initial_state_fn=networks.initial_state_fn,
        config=config,
        counter=counter,
        logger=logger,
    )



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SVG learner implementation."""

import time
from typing import Dict, Iterator, List, Optional

import acme
from acme.agents.tf.svg0_prior import utils as svg0_utils
from acme.tf import savers as tf2_savers
from acme.tf import utils as tf2_utils
from acme.utils import counting
from acme.utils import loggers
import numpy as np
import reverb
import sonnet as snt
import tensorflow as tf
from trfl import continuous_retrace_ops

_MIN_LOG_VAL = 1e-20


class SVG0Learner(acme.Learner):
  """SVG0 learner with optional prior.

  This is the learning component of an SVG0 agent. IE it takes a dataset as
  input and implements update functionality to learn from this dataset.
  """

  def __init__(
      self,
      policy_network: snt.Module,
      critic_network: snt.Module,
      target_policy_network: snt.Module,
      target_critic_network: snt.Module,
      discount: float,
      target_update_period: int,
      dataset_iterator: Iterator[reverb.ReplaySample],
      prior_network: Optional[snt.Module] = None,
      target_prior_network: Optional[snt.Module] = None,
      policy_optimizer: Optional[snt.Optimizer] = None,
      critic_optimizer: Optional[snt.Optimizer] = None,
      prior_optimizer: Optional[snt.Optimizer] = None,
      distillation_cost: Optional[float] = 1e-3,
      entropy_regularizer_cost: Optional[float] = 1e-3,
      num_action_samples: int = 10,
      lambda_: float = 1.0,
      counter: Optional[counting.Counter] = None,
      logger: Optional[loggers.Logger] = None,
      checkpoint: bool = True,
  ):
    """Initializes the learner.

    Args:
      policy_network: the online (optimized) policy.
      critic_network: the online critic.
      target_policy_network: the target policy (which lags behind the online
        policy).
      target_critic_network: the target critic.
      discount: discount to use for TD updates.
      target_update_period: number of learner steps to perform before updating
        the target networks.
      dataset_iterator: dataset to learn from, whether fixed or from a replay
        buffer (see `acme.datasets.reverb.make_dataset` documentation).
      prior_network: the online (optimized) prior.
      target_prior_network: the target prior (which lags behind the online
        prior).
      policy_optimizer: the optimizer to be applied to the SVG-0 (policy) loss.
      critic_optimizer: the optimizer to be applied to the distributional
        Bellman loss.
      prior_optimizer: the optimizer to be applied to the prior (distillation)
        loss.
      distillation_cost: a multiplier to be used when adding distillation
        against the prior to the losses.
      entropy_regularizer_cost: a multiplier used for per state sample based
        entropy added to the actor loss.
      num_action_samples: the number of action samples to use for estimating the
        value function and sample based entropy.
      lambda_: the `lambda` value to be used with retrace.
      counter: counter object used to keep track of steps.
      logger: logger object to be used by learner.
      checkpoint: boolean indicating whether to checkpoint the learner.
    """

    # Store online and target networks.
    self._policy_network = policy_network
    self._critic_network = critic_network
    self._target_policy_network = target_policy_network
    self._target_critic_network = target_critic_network

    self._prior_network = prior_network
    self._target_prior_network = target_prior_network

    self._lambda = lambda_
    self._num_action_samples = num_action_samples
    self._distillation_cost = distillation_cost
    self._entropy_regularizer_cost = entropy_regularizer_cost

    # General learner book-keeping and loggers.
    self._counter = counter or counting.Counter()
    self._logger = logger or loggers.make_default_logger('learner')

    # Other learner parameters.
    self._discount = discount

    # Necessary to track when to update target networks.
    self._num_steps = tf.Variable(0, dtype=tf.int32)
    self._target_update_period = target_update_period

    # Batch dataset and create iterator.
    self._iterator = dataset_iterator

    # Create optimizers if they aren't given.
    self._critic_optimizer = critic_optimizer or snt.optimizers.Adam(1e-4)
    self._policy_optimizer = policy_optimizer or snt.optimizers.Adam(1e-4)
    self._prior_optimizer = prior_optimizer or snt.optimizers.Adam(1e-4)

    # Expose the variables.
    self._variables = {
        'critic': self._critic_network.variables,
        'policy': self._policy_network.variables,
    }
    if self._prior_network is not None:
      self._variables['prior'] = self._prior_network.variables

    # Create a checkpointer and snapshotter objects.
    self._checkpointer = None
    self._snapshotter = None

    if checkpoint:
      objects_to_save = {
          'counter': self._counter,
          'policy': self._policy_network,
          'critic': self._critic_network,
          'target_policy': self._target_policy_network,
          'target_critic': self._target_critic_network,
          'policy_optimizer': self._policy_optimizer,
          'critic_optimizer': self._critic_optimizer,
          'num_steps': self._num_steps,
      }
      if self._prior_network is not None:
        objects_to_save['prior'] = self._prior_network
        objects_to_save['target_prior'] = self._target_prior_network
        objects_to_save['prior_optimizer'] = self._prior_optimizer

      self._checkpointer = tf2_savers.Checkpointer(
          subdirectory='svg0_learner',
          objects_to_save=objects_to_save)
      objects_to_snapshot = {
          'policy': self._policy_network,
          'critic': self._critic_network,
      }
      if self._prior_network is not None:
        objects_to_snapshot['prior'] = self._prior_network

      self._snapshotter = tf2_savers.Snapshotter(
          objects_to_save=objects_to_snapshot)

    # Do not record timestamps until after the first learning step is done.
    # This is to avoid including the time it takes for actors to come online and
    # fill the replay buffer.
    self._timestamp = None

  @tf.function
  def _step(self) -> Dict[str, tf.Tensor]:
    # Update target network
    online_variables = [
        *self._critic_network.variables,
        *self._policy_network.variables,
    ]
    if self._prior_network is not None:
      online_variables += [*self._prior_network.variables]
    online_variables = tuple(online_variables)

    target_variables = [
        *self._target_critic_network.variables,
        *self._target_policy_network.variables,
    ]
    if self._prior_network is not None:
      target_variables += [*self._target_prior_network.variables]
    target_variables = tuple(target_variables)

    # Make online -> target network update ops.
    if tf.math.mod(self._num_steps, self._target_update_period) == 0:
      for src, dest in zip(online_variables, target_variables):
        dest.assign(src)
    self._num_steps.assign_add(1)

    # Get data from replay (dropping extras if any) and flip to `[T, B, ...]`.
    sample: reverb.ReplaySample = next(self._iterator)
    data = tf2_utils.batch_to_sequence(sample.data)
    observations, actions, rewards, discounts, extra = (data.observation,
                                                        data.action,
                                                        data.reward,
                                                        data.discount,
                                                        data.extras)
    online_target_pi_q = svg0_utils.OnlineTargetPiQ(
        online_pi=self._policy_network,
        online_q=self._critic_network,
        target_pi=self._target_policy_network,
        target_q=self._target_critic_network,
        num_samples=self._num_action_samples,
        online_prior=self._prior_network,
        target_prior=self._target_prior_network,
    )
    with tf.GradientTape(persistent=True) as tape:
      step_outputs = svg0_utils.static_rnn(
          core=online_target_pi_q,
          inputs=(observations, actions),
          unroll_length=rewards.shape[0])

      # Flip target samples to have shape [S, T+1, B, ...] where 'S' is the
      # number of action samples taken.
      target_pi_samples = tf2_utils.batch_to_sequence(
          step_outputs.target_samples)
      # Tile observations to have shape [S, T+1, B,..].
      tiled_observations = tf2_utils.tile_nested(observations,
                                                 self._num_action_samples)

      # Finally compute target Q values on the new action samples.
      # Shape: [S, T+1, B, 1]
      target_q_target_pi_samples = snt.BatchApply(self._target_critic_network,
                                                  3)(tiled_observations,
                                                     target_pi_samples)
      # Compute the value estimate by averaging over the action dimension.
      # Shape: [T+1, B, 1].
      target_v_target_pi = tf.reduce_mean(target_q_target_pi_samples, axis=0)

      # Split the target V's into the target for learning
      # `value_function_target` and the bootstrap value. Shape: [T, B].
      value_function_target = tf.squeeze(target_v_target_pi[:-1], axis=-1)
      # Shape: [B].
      bootstrap_value = tf.squeeze(target_v_target_pi[-1], axis=-1)

      # When learning with a prior, add entropy terms to value targets.
      if self._prior_network is not None:
        value_function_target -= self._distillation_cost * tf.stop_gradient(
            step_outputs.analytic_kl_to_target[:-1]
            )
        bootstrap_value -= self._distillation_cost * tf.stop_gradient(
            step_outputs.analytic_kl_to_target[-1])

      # Get target log probs and behavior log probs from rollout.
      # Shape: [T+1, B].
      target_log_probs_behavior_actions = (
          step_outputs.target_log_probs_behavior_actions)
      behavior_log_probs = extra['log_prob']
      # Calculate importance weights. Shape: [T+1, B].
      rhos = tf.exp(target_log_probs_behavior_actions - behavior_log_probs)

      # Filter the importance weights to mask out episode restarts. Ignore the
      # last action and consider the step type of the next step for masking.
      # Shape: [T, B].
      episode_start_mask = tf2_utils.batch_to_sequence(
          sample.data.start_of_episode)[1:]

      rhos = svg0_utils.mask_out_restarting(rhos[:-1], episode_start_mask)

      # rhos = rhos[:-1]
      # Compute the log importance weights with a small value added for
      # stability.
      # Shape: [T, B]
      log_rhos = tf.math.log(rhos + _MIN_LOG_VAL)

      # Retrieve the target and online Q values and throw away the last action.
      # Shape: [T, B].
      target_q_values = tf.squeeze(step_outputs.target_q[:-1], -1)
      online_q_values = tf.squeeze(step_outputs.online_q[:-1], -1)

      # Flip target samples to have shape [S, T+1, B, ...] where 'S' is the
      # number of action samples taken.
      online_pi_samples = tf2_utils.batch_to_sequence(
          step_outputs.online_samples)
      target_q_online_pi_samples = snt.BatchApply(self._target_critic_network,
                                                  3)(tiled_observations,
                                                     online_pi_samples)
      expected_q = tf.reduce_mean(
          tf.squeeze(target_q_online_pi_samples, -1), axis=0)

      # Flip online_log_probs to be of shape [S, T+1, B] and then compute
      # entropy by averaging over num samples. Final shape: [T+1, B].
      online_log_probs = tf2_utils.batch_to_sequence(
          step_outputs.online_log_probs)
      sample_based_entropy = tf.reduce_mean(-online_log_probs, axis=0)
      retrace_outputs = continuous_retrace_ops.retrace_from_importance_weights(
          log_rhos=log_rhos,
          discounts=self._discount * discounts[:-1],
          rewards=rewards[:-1],
          q_values=target_q_values,
          values=value_function_target,
          bootstrap_value=bootstrap_value,
          lambda_=self._lambda,
      )

      # Critic loss. Shape: [T, B].
      critic_loss = 0.5 * tf.math.squared_difference(
          tf.stop_gradient(retrace_outputs.qs), online_q_values)

      # Policy loss- SVG0 with sample based entropy. Shape: [T, B]
      policy_loss = -(
          expected_q + self._entropy_regularizer_cost * sample_based_entropy)
      policy_loss = policy_loss[:-1]

      if self._prior_network is not None:
        # When training the prior, also add the per-timestep KL cost.
        policy_loss += (
            self._distillation_cost * step_outputs.analytic_kl_to_target[:-1])

      # Ensure episode restarts are masked out when computing the losses.
      critic_loss = svg0_utils.mask_out_restarting(critic_loss,
                                                   episode_start_mask)
      critic_loss = tf.reduce_mean(critic_loss)

      policy_loss = svg0_utils.mask_out_restarting(policy_loss,
                                                   episode_start_mask)
      policy_loss = tf.reduce_mean(policy_loss)

      if self._prior_network is not None:
        prior_loss = step_outputs.analytic_kl_divergence[:-1]
        prior_loss = svg0_utils.mask_out_restarting(prior_loss,
                                                    episode_start_mask)
        prior_loss = tf.reduce_mean(prior_loss)

    # Get trainable variables.
    policy_variables = self._policy_network.trainable_variables
    critic_variables = self._critic_network.trainable_variables

    # Compute gradients.
    policy_gradients = tape.gradient(policy_loss, policy_variables)
    critic_gradients = tape.gradient(critic_loss, critic_variables)
    if self._prior_network is not None:
      prior_variables = self._prior_network.trainable_variables
      prior_gradients = tape.gradient(prior_loss, prior_variables)

    # Delete the tape manually because of the persistent=True flag.
    del tape

    # Apply gradients.
    self._policy_optimizer.apply(policy_gradients, policy_variables)
    self._critic_optimizer.apply(critic_gradients, critic_variables)
    losses = {
        'critic_loss': critic_loss,
        'policy_loss': policy_loss,
    }

    if self._prior_network is not None:
      self._prior_optimizer.apply(prior_gradients, prior_variables)
      losses['prior_loss'] = prior_loss

    # Losses to track.
    return losses

  def step(self):
    # Run the learning step.
    fetches = self._step()

    # Compute elapsed time.
    timestamp = time.time()
    elapsed_time = timestamp - self._timestamp if self._timestamp else 0
    self._timestamp = timestamp

    # Update our counts and record it.
    counts = self._counter.increment(steps=1, walltime=elapsed_time)
    fetches.update(counts)

    # Checkpoint and attempt to write the logs.
    if self._checkpointer is not None:
      self._checkpointer.save()
    if self._snapshotter is not None:
      self._snapshotter.save()
    self._logger.write(fetches)

  def get_variables(self, names: List[str]) -> List[List[np.ndarray]]:
    return [tf2_utils.to_numpy(self._variables[name]) for name in names]



# python3
# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Simple JAX actors."""

from typing import Generic, Optional

from acme import adders
from acme import core
from acme import types
from acme.agents.jax import actor_core
from acme.jax import networks as network_lib
from acme.jax import utils
from acme.jax import variable_utils
import dm_env
import jax


class GenericActor(core.Actor, Generic[actor_core.State, actor_core.Extras]):
  """A generic actor implemented on top of ActorCore.

  An actor based on a policy which takes observations and outputs actions. It
  also adds experiences to replay and updates the actor weights from the policy
  on the learner.
  """

  def __init__(
      self,
      actor: actor_core.ActorCore[actor_core.State, actor_core.Extras],
      random_key: network_lib.PRNGKey,
      variable_client: Optional[variable_utils.VariableClient],
      adder: Optional[adders.Adder] = None,
      jit: bool = True,
      backend: Optional[str] = 'cpu',
      per_episode_update: bool = False
  ):
    """Initializes a feed forward actor.

    Args:
      actor: actor core.
      random_key: Random key.
      variable_client: The variable client to get policy parameters from.
      adder: An adder to add experiences to.
      jit: Whether or not to jit the passed ActorCore's pure functions.
      backend: Which backend to use when jitting the policy.
      per_episode_update: if True, updates variable client params once at the
        beginning of each episode
    """
    self._random_key = random_key
    self._variable_client = variable_client
    self._adder = adder
    self._state = None

    # Unpack ActorCore, jitting if requested.
    if jit:
      self._init = jax.jit(actor.init, backend=backend)
      self._policy = jax.jit(actor.select_action, backend=backend)
    else:
      self._init = actor.init
      self._policy = actor.select_action
    self._get_extras = actor.get_extras
    self._per_episode_update = per_episode_update

  @property
  def _params(self):
    return self._variable_client.params if self._variable_client else []

  def select_action(self,
                    observation: network_lib.Observation) -> types.NestedArray:
    action, self._state = self._policy(self._params, observation, self._state)
    return utils.to_numpy(action)

  def observe_first(self, timestep: dm_env.TimeStep):
    self._random_key, key = jax.random.split(self._random_key)
    self._state = self._init(key)
    if self._adder:
      self._adder.add_first(timestep)
    if self._variable_client and self._per_episode_update:
      self._variable_client.update_and_wait()

  def observe(self, action: network_lib.Action, next_timestep: dm_env.TimeStep):
    if self._adder:
      self._adder.add(
          action, next_timestep, extras=self._get_extras(self._state))

  def update(self, wait: bool = False):
    if self._variable_client and not self._per_episode_update:
      self._variable_client.update(wait)
